import logging

import pandas as pd


model = "opendrift"
logger = logging.getLogger(model)


def check_plots(which_plots, export_variables, drift_model):
    """Check that input plot options are valid,

    particularly that the necessary export variables are included.

    Parameters
    ----------
    which_plots : dict
        Dictionary of plot options.
    export_variables : list
        Variables to be saved to file.
    drift_model : str
        Which OpenDrift model is being used.

    Raises
    ------
    ValueError
        If "all" is specified with other plot options.
    ValueError
        If necessary export variables are missing.
    ValueError
        If oil budget plot is requested with a drift model other than OpenOil.
    """

    if (
        len([k for k in which_plots.keys() if "oil" in k]) > 0
        and drift_model != "OpenOil"
    ):
        raise ValueError("Oil budget plot only available for OpenOil drift model")

    # if "all", make sure it is the only plot option
    # since it will create its own dictionary of plot options
    if "all" in which_plots and len(which_plots) > 1:
        raise ValueError(
            "If 'all' is specified for plots, it must be the only plot option."
        )

    # check for cases that require an export variable
    if export_variables is not None:

        # flatten dict
        which_plots_flat = pd.json_normalize(which_plots).to_dict(orient="records")

        if len(which_plots_flat) == 0:
            return

        which_plots_flat = which_plots_flat[0]

        plot_options_to_check = [
            "linecolor",
            "color",
            "background",
            "variable",
            "markersize",
        ]
        missing_variables = []
        for plot_option in plot_options_to_check:
            variables = [
                v
                for k, v in which_plots_flat.items()
                if plot_option == k.split(".")[1] and v not in export_variables
            ]
            missing_variables.extend(variables)

        # handle oil budget separately
        vars_needed = [
            "x_wind",
            "y_wind",
            "x_sea_water_velocity",
            "y_sea_water_velocity",
        ]
        variables = [
            list(set(vars_needed) - set(export_variables))
            for k, v in which_plots_flat.items()
            if "show_wind_and_current" in k
            and v
            and (len(set(vars_needed) - set(export_variables)) > 0)
        ]
        if len(variables) > 0:
            variables = variables[0]
            missing_variables.extend(variables)

        if len(missing_variables) > 0:
            raise ValueError(
                f"Missing export variables for the following plot options: {list(set(missing_variables))}"
            )


def make_filename_string(plot_name, filename, kwargs):
    """Create a filename string based on the kwargs."""

    # modify filename with plot kwargs
    if len(kwargs) > 0:
        filename += "_" + "_".join(
            [
                f"{key}_{value}"
                for key, value in kwargs.items()
                if key not in ["filetype"]
            ]
        )

    # include filetype suffix
    if "spaghetti" in plot_name or "oil" in plot_name or "property" in plot_name:
        if "filetype" not in kwargs:
            filename += ".png"
        else:
            filename += "." + kwargs["filetype"]
            kwargs.pop("filetype")

    elif "animation" in plot_name:
        if "filetype" not in kwargs:
            filename += ".gif"
        else:
            filename += "." + kwargs["filetype"]
            kwargs.pop("filetype")

    else:
        raise ValueError(f"Invalid plot name: {plot_name}")

    return filename


def plot(plot_name, input_kwargs, o, filename, drift_model):
    """Create a plot based on the plot name and input kwargs.

    Parameters
    ----------
    plot_name : str
        Must contain key word to identify the plot type.
    input_kwargs : dict
        Optional plot kwargs.
    o : _type_
        OpenDrift simulation object.
    filename : str
        Filename substring to save the plot.
    drift_model : str
        Which OpenDrift model is being used.
    """

    if "spaghetti" in plot_name or plot_name == "all":

        filename += "_spaghetti"

        # add input plot input_kwargs to the default kwargs
        kwargs = {"fast": True}
        kwargs.update(input_kwargs)

        # modify filename with plot kwargs
        filename = make_filename_string(plot_name, filename, kwargs)

        o.plot(filename=filename, **kwargs)

    elif (
        "animation" in plot_name and "profile" not in plot_name
    ) or plot_name == "all":

        # add input plot kwargs to the default kwargs
        kwargs = {"fast": True, "fps": 4}
        kwargs.update(input_kwargs)

        # modify filename with plot kwargs
        filename = make_filename_string(plot_name, filename, kwargs)

        o.animation(filename=filename, **kwargs)

    elif ("animation" in plot_name and "profile" in plot_name) or plot_name == "all":

        filename += "_profile"

        # add input plot kwargs to the default kwargs
        kwargs = {"fps": 4}
        kwargs.update(input_kwargs)

        # modify filename with plot kwargs
        filename = make_filename_string(plot_name, filename, kwargs)

        o.animation_profile(filename=filename, **kwargs)

    elif "oil" in plot_name or plot_name == "all":

        filename += "_oil"
        plot_name = "oil"

        # add input plot kwargs to the default kwargs
        kwargs = {}
        kwargs = {
            "show_wind_and_current": True,
            "show_watercontent_and_viscosity": True,
        }
        kwargs.update(input_kwargs)

        # modify filename with plot kwargs
        filename = make_filename_string(plot_name, filename, kwargs)

        if drift_model == "OpenOil":
            o.plot_oil_budget(filename=filename, **kwargs)
        else:
            raise ValueError("Oil budget plot only available for OpenOil drift model")

    elif "property" in plot_name:  # or plot_name == "all":

        if not "variable" in input_kwargs:
            raise ValueError(
                "Property plot must be specified as 'variable' in the plots dictionary."
            )

        filename += "_property"
        plot_name = "property"

        # add input plot kwargs to the default kwargs
        kwargs = {}
        kwargs.update(input_kwargs)

        # modify filename with plot kwargs
        filename = make_filename_string(plot_name, filename, kwargs)

        o.plot_property(filename=filename, **kwargs)

    logger.info(f"Saved plot to {filename}")

    return filename


def make_plots(which_plots, o, filename, drift_model):
    """Run through each plot key and make the plot.

    This extra level of abstraction is necessary to allow for multiple
    types of each plot to be made, for example, a spaghetti plot and a
    spaghetti plot with tracks colored.
    """

    if "all" in which_plots:
        which_plots = {
            "spaghetti": {},
            "animation": {},
            "animation_profile": {},
        }
        if drift_model == "OpenOil":
            which_plots["oil"] = {}

    for which_plot in which_plots:
        filename_out = plot(
            which_plot, which_plots[which_plot], o, filename, drift_model
        )
        # save filename for plot
        which_plots[which_plot]["filename"] = filename_out
    return which_plots


def make_plots_after_simulation(output_filepath, plots="all"):
    """Make plots after a simulation has been run.

    Parameters
    ----------
    output_filepath : str
        Path to the output file from the simulation.
    which_plots : dict
        Dictionary of plot options.

    Returns
    -------
    dict
        Dictionary of plot options with the filename of each plot.
    """
    import opendrift as od

    # load output file
    o = od.open(str(output_filepath))

    # want output_file to not include any suffix
    output_filepath = (
        str(output_filepath)
        .replace(".nc", "")
        .replace(".parquet", "")
        .replace(".parq", "")
    )

    # figure out drift_model
    drift_model = type(o).__name__

    # make plots
    plots = make_plots(plots, o, output_filepath, drift_model)

    return plots



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import pathlib
import sys

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# see https://pypi.org/project/setuptools-scm/ for details
from importlib.metadata import version as imversion


print("python exec:", sys.executable)
print("sys.path:", sys.path)
root = pathlib.Path(__file__).parent.parent.absolute()
os.environ["PYTHONPATH"] = str(root)
sys.path.insert(0, str(root))

import particle_tracking_manager  # isort:skip

# -- Project information -----------------------------------------------------

project = "particle-tracking-manager"
copyright = "2023-, axiom-data-science"
author = "axiom-data-science"

release = imversion("particle-tracking-manager")
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    # "nbsphinx",
    "recommonmark",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "numpydoc",
    # "nbsphinx",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinxcontrib.srclinks",
    "myst_nb",
]

myst_enable_extensions = [
    "dollarmath",  # enables $...$ and $$...$$
    "amsmath",  # enables \begin{equation}, \text{}, etc.
]

# for compiling notebooks with mystnb
# https://docs.readthedocs.io/en/stable/guides/jupyter.html#using-notebooks-in-other-formats
nb_custom_formats = {
    ".md": ["jupytext.reads", {"fmt": "mystnb"}],
}

# https://myst-nb.readthedocs.io/en/v0.9.0/use/execute.html
# jupyter_execute_notebooks = "off"
nb_execution_mode = "cache"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "*.ipynb",
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = "particle-tracking-manager documentation"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# https://myst-nb.readthedocs.io/en/v0.13.0/use/execute.html#execution-timeout
# had this message:
# WARNING: 'execution_timeout' is deprecated for 'nb_execution_timeout' [mystnb.config]
# WARNING: 'execution_allow_errors' is deprecated for 'nb_execution_allow_errors' [mystnb.config]
nb_execution_timeout = 180  # seconds.


# -- nbsphinx specific options ----------------------------------------------
# this allows notebooks to be run even if they produce errors.
nbsphinx_allow_errors = True


# copied from cf-xarray
# autosummary_generate = True

autodoc_typehints = "none"
autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "show-inheritance": True,
    "undoc-members": True,
    "private-members": True,
}
napoleon_use_param = True
napoleon_use_rtype = True



"""Enums for OpenDrift."""

from .object_types import ObjectTypeEnum
from .oil_types import NAME_TO_OIL_ID, OIL_ID_TO_NAME, OilTypeEnum
from .others import (
    CoastlineActionEnum,
    DiffusivityModelEnum,
    DriftModelEnum,
    DropletSizeDistributionEnum,
    PlotTypeEnum,
    RadiusTypeEnum,
    SeafloorActionEnum,
)
from .species_types import HABSpeciesTypeEnum


__all__ = [
    "OilTypeEnum",
    "OIL_ID_TO_NAME",
    "NAME_TO_OIL_ID",
    "ObjectTypeEnum",
    "DriftModelEnum",
    "RadiusTypeEnum",
    "DiffusivityModelEnum",
    "CoastlineActionEnum",
    "SeafloorActionEnum",
    "HABSpeciesTypeEnum",
    "PlotTypeEnum",
    "DropletSizeDistributionEnum",
]



"""Test realistic scenarios, which are slower."""

import pickle

from pathlib import Path

import numpy as np
import pytest
import xarray as xr

import particle_tracking_manager as ptm


def is_netcdf(path):
    with open(path, "rb") as f:
        sig = f.read(8)
    return sig.startswith(b"CDF") or sig.startswith(b"\x89HDF\r\n\x1a\n")


def is_parquet(path):
    with open(path, "rb") as f:
        start = f.read(4)
        f.seek(-4, 2)
        end = f.read(4)
    return start == b"PAR1" and end == b"PAR1"


# set up an alternate dataset on-the-fly
ds = xr.Dataset(
    data_vars={
        "u": (("ocean_time", "Z", "Y", "X"), np.ones((2, 3, 4, 5))),
        "v": (("ocean_time", "Z", "Y", "X"), np.ones((2, 3, 4, 5))),
        "w": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 4, 5))),
        "salt": (("ocean_time", "Z", "Y", "X"), np.ones((2, 3, 4, 5)) * 31),
        "temp": (("ocean_time", "Z", "Y", "X"), np.ones((2, 3, 4, 5)) * 18),
        "wetdry_mask_rho": (("ocean_time", "Y", "X"), np.ones((2, 4, 5))),
        "mask_rho": (("Y", "X"), np.ones((4, 5))),
        "h": (("Y", "X"), np.ones((4, 5)) * 10),
        "angle": (("Y", "X"), np.zeros((4, 5))),
        "Uwind": (("ocean_time", "Y", "X"), np.zeros((2, 4, 5))),
        "Vwind": (("ocean_time", "Y", "X"), np.zeros((2, 4, 5))),
        "Cs_r": (("Z"), np.linspace(-1, 0, 3)),
        "hc": 16,
    },
    coords={
        # "ocean_time": ("ocean_time", ["1970-01-01T00:00:00", "1970-01-01T00:10:00"], {"units": "seconds since 1970-01-01"}),
        "ocean_time": (
            "ocean_time",
            [0, 60 * 10],
            {"units": "seconds since 1970-01-01"},
        ),
        "s_rho": (("Z"), np.linspace(-1, 0, 3)),
        "lon_rho": (
            ("Y", "X"),
            np.array(
                [
                    [1, 1.5, 2, 2.5, 3],
                    [1, 1.5, 2, 2.5, 3],
                    [1, 1.5, 2, 2.5, 3],
                    [1, 1.5, 2, 2.5, 3],
                ]
            ),
        ),
        "lat_rho": (
            ("Y", "X"),
            np.array(
                [
                    [1, 1.25, 1.5, 1.75, 2],
                    [1, 1.25, 1.5, 1.75, 2],
                    [1, 1.25, 1.5, 1.75, 2],
                    [1, 1.25, 1.5, 1.75, 2],
                ]
            ),
        ),
    },
)
ds_info = dict(
    lon_min=1,
    lon_max=3,
    lat_min=1,
    lat_max=2,
    start_time_model=0,
    end_time_fixed=60 * 10,
)

ptm.config_ocean_model.register_on_the_fly(ds_info)


# also to use the user-defined template of the TXLA model, need to input where pooch is downloading
# the file
ptm.config_ocean_model.update_TXLA_with_download_location()


@pytest.mark.slow
def test_add_new_reader():
    """Add a separate reader from the defaults using ds."""

    manager = ptm.OpenDriftModel(
        steps=1, ocean_model="ONTHEFLY", lon=1.2, lat=1.2, start_time=0, time_step=0.01
    )
    manager.add_reader(ds=ds)


@pytest.mark.slow
def test_run_netcdf_and_plot():
    """Set up and run."""

    import tempfile

    ts = 6 * 60  # 6 minutes in seconds

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        manager = ptm.OpenDriftModel(
            **seeding_kwargs,
            use_static_masks=True,
            steps=2,
            output_format="netcdf",
            use_cache=True,
            interpolator_filename=temp_file.name,
            ocean_model="TXLA",
            ocean_model_local=False,
            plots={
                "all": {},
            },
            time_step=ts,
        )
        manager.run_all()

        assert "nc" in manager.o.outfile_name
        assert manager.config.interpolator_filename == Path(temp_file.name).with_suffix(
            ".pickle"
        )

        # Replace 'path_to_pickle_file.pkl' with the actual path to your pickle file
        with open(manager.config.interpolator_filename, "rb") as file:
            data = pickle.load(file)
        assert "spl_x" in data
        assert "spl_y" in data

    # check time_step across access points
    assert (
        # m.o._config["general:time_step_minutes"]["value"]  # this is not correct, don't know why
        manager.o.time_step.total_seconds()
        == ts
        == manager.config.time_step
        # == m.o.get_configspec()["general:time_step_minutes"]["value"]  # this is not correct, don't know why
    )


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_biomass_change():
    """Set up and run HarmfulAlgalBloom and match biomass change."""

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
    )
    m.add_reader()

    # change model values for test
    inds = m.ds.temp.notnull()
    m.ds["temp"].values[inds] = 15.0
    m.ds["salt"].values[inds] = 32.0
    m.run_all()

    # check that biomass decreased due to temperature-induced mortality
    # calculated as: biomass = initial_biomass * exp(growth_rate-mortality_rate_high * time)
    assert np.allclose(
        float(m.o.elements.biomass[0]),
        np.exp(
            (m.config.growth_rate_high - m.config.mortality_rate_low) * 3600 / 86400
        ),
    )


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_vertical_behavior_band():
    """Set up and run HarmfulAlgalBloom and match vertical change."""

    seeding_kwargs = dict(
        lon=-90,
        lat=28.7,
        number=1,
        start_time="2009-11-19T12:00:00",
        swim_speed=0.001,
        z=-30,
    )
    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
        vertical_behavior="band",
        band_center_depth=-10.0,
        band_half_width=0.0,
        vertical_mixing=False,
        do3D=False,
    )
    m.add_reader()
    m.run_all()

    dz = seeding_kwargs["swim_speed"] * 3600  # swim speed * time in seconds
    assert np.allclose(float(m.o.elements.z[0]), seeding_kwargs["z"] + dz)


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_vertical_behavior_diel_band():
    """Set up and run HarmfulAlgalBloom with diel_band and match vertical change.

    Particle starts deeper than both diel_day_depth and diel_night_depth, so
    it should swim upward at constant swim_speed for the duration, independent
    of whether this hour is classified as day or night.
    """

    seeding_kwargs = dict(
        lon=-90.0,
        lat=28.7,
        number=1,
        start_time="2009-11-19T12:00:00",  # same as band test
        swim_speed=0.001,
        z=-50.0,
    )

    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
        vertical_behavior="diel_band",
        diel_day_depth=-20.0,
        diel_night_depth=-10.0,
        vertical_mixing=False,
        do3D=False,
    )
    m.add_reader()
    m.run_all()

    # Analytical expectation: no vertical advection/mixing, so only active
    # swimming contributes. 1 hour = 3600 s.
    dz = seeding_kwargs["swim_speed"] * 3600  # swim_speed * time (s)
    expected_z = seeding_kwargs["z"] + dz

    assert np.allclose(float(m.o.elements.z[0]), expected_z)


# reinstitute this test once OpenDrift PR is accepted that outputs parquet files directly
# @pytest.mark.slow
# def test_run_parquet():
#     """Set up and run."""

#     seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
#     manager = ptm.OpenDriftModel(
#         **seeding_kwargs,
#         use_static_masks=True,
#         steps=2,
#         output_format="parquet",
#         ocean_model="TXLA",
#         ocean_model_local=False,
#     )
#     manager.run_all()

#     assert "parquet" in manager.o.outfile_name


@pytest.mark.slow
def test_run_parquet_and_netcdf():
    """Set up and run."""

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00")
    manager = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        steps=2,
        output_format="both",
        ocean_model="TXLA",
        ocean_model_local=False,
    )
    manager.run_all()

    assert "nc" in manager.o.outfile_name

    # 2. parquet file with same stem exists
    out_parquet = Path(manager.o.outfile_name).with_suffix(".parquet")
    assert out_parquet.exists()

    # Check actual file format signatures
    assert is_netcdf(manager.o.outfile_name), "NC file is not valid netCDF"
    assert not is_parquet(
        manager.o.outfile_name
    ), "NC file is incorrectly a parquet file"

    assert is_parquet(out_parquet), "Parquet file is not valid parquet"
    assert not is_netcdf(out_parquet), "Parquet file is incorrectly netCDF"


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_PN():
    """Set up and run HarmfulAlgalBloom for Pseudo Nitzschia."""

    seeding_kwargs = dict(
        lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00", z=-20
    )
    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
        species_type="PN",
        vertical_mixing=False,
        do3D=False,
    )
    m.add_reader()

    # change model values for test
    inds = m.ds.temp.notnull()
    m.ds["temp"].values[inds] = m.config.temperature_pref_min
    m.ds["salt"].values[inds] = (
        m.config.salinity_death_min + 0.5
    )  # this causes medium mortality/growth
    m.run_all()

    # check that biomass decreased due to temperature-induced mortality
    # calculated as: biomass = initial_biomass * exp(growth_rate-mortality_rate_high * time)
    assert np.allclose(
        float(m.o.elements.biomass[0]),
        np.exp(
            (m.config.growth_rate_medium - m.config.mortality_rate_medium)
            * 3600
            / 86400
        ),
    )

    # Analytical expectation: no vertical advection/mixing, so only active
    # swimming contributes. 1 hour = 3600 s.
    # particles are swimming downward from near the surface
    dz = m.config.swim_speed * 3600  # swim_speed * time (s)
    expected_z = seeding_kwargs["z"] - dz
    assert np.allclose(float(m.o.elements.z[0]), expected_z)


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_AX():
    """Set up and run HarmfulAlgalBloom for Alexandrium."""

    seeding_kwargs = dict(
        lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00", z=-40
    )
    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
        species_type="AX",
        vertical_mixing=False,
        do3D=False,
    )
    m.add_reader()

    # change model values for test
    inds = m.ds.temp.notnull()
    m.ds["temp"].values[inds] = (
        m.config.temperature_death_max + 0.5
    )  # this causes high mortality/low growth
    m.ds["salt"].values[inds] = m.config.salinity_pref_max
    m.run_all()

    # check that biomass decreased due to temperature-induced mortality
    # calculated as: biomass = initial_biomass * exp(growth_rate-mortality_rate_high * time)
    assert np.allclose(
        float(m.o.elements.biomass[0]),
        np.exp(
            (m.config.growth_rate_low - m.config.mortality_rate_high) * 3600 / 86400
        ),
    )

    # Analytical expectation: no vertical advection/mixing, so only active
    # swimming contributes. 1 hour = 3600 s.
    dz = m.config.swim_speed * 3600  # swim_speed * time (s)
    expected_z = seeding_kwargs["z"] + dz
    assert np.allclose(float(m.o.elements.z[0]), expected_z)


@pytest.mark.slow
def test_run_HarmfulAlgalBloom_DP():
    """Set up and run HarmfulAlgalBloom for Dinophysis."""

    seeding_kwargs = dict(
        lon=-90, lat=28.7, number=1, start_time="2009-11-19T12:00:00", z=-40
    )
    m = ptm.OpenDriftModel(
        **seeding_kwargs,
        use_static_masks=True,
        duration="1h",
        ocean_model="TXLA",
        ocean_model_local=False,
        drift_model="HarmfulAlgalBloom",
        species_type="DP",
        vertical_mixing=False,
        do3D=False,
    )
    m.add_reader()

    # change model values for test
    inds = m.ds.temp.notnull()
    m.ds["temp"].values[inds] = m.config.temperature_pref_max
    m.ds["salt"].values[inds] = m.config.salinity_pref_max
    m.run_all()

    # check that biomass decreased due to temperature-induced mortality
    # calculated as: biomass = initial_biomass * exp(growth_rate-mortality_rate_high * time)
    assert np.allclose(
        float(m.o.elements.biomass[0]),
        np.exp(
            (m.config.growth_rate_high - m.config.mortality_rate_low) * 3600 / 86400
        ),
    )

    # Analytical expectation: no vertical advection/mixing, so only active
    # swimming contributes. 1 hour = 3600 s.
    dz = m.config.swim_speed * 3600  # swim_speed * time (s)
    expected_z = seeding_kwargs["z"] + dz
    assert np.allclose(float(m.o.elements.z[0]), expected_z)



"""Command line interface to get inputs from web application."""

# Standard library imports
import argparse
import ast

from datetime import datetime

# Third-party imports
import pandas as pd

# Local imports
import particle_tracking_manager as ptm


def is_int(s: str) -> bool:
    """Check if string is actually int."""
    try:
        int(s)
        return True
    except (ValueError, TypeError):
        return False


def is_float(s: str) -> bool:
    """Check if string is actually float."""
    try:
        float(s)
        return True
    except (ValueError, TypeError):
        return False


def is_None(s: str) -> bool:
    """Check if string is actually None."""
    if s == "None":
        return True
    else:
        return False


def is_datestr(s: str) -> bool:
    """Check if string is actually a datestring."""

    try:
        out = pd.Timestamp(s)
        assert not pd.isnull(out)
        return True
    except (ValueError, AssertionError, TypeError):
        return False


def is_deltastr(s: str) -> bool:
    """Check if string is actually a Timedelta."""

    try:
        out = pd.Timedelta(s)
        assert not pd.isnull(out)
        return True
    except (ValueError, AssertionError):
        return False


# https://sumit-ghosh.com/articles/parsing-dictionary-key-value-pairs-kwargs-argparse-python/
class ParseKwargs(argparse.Action):
    """With can user can input dicts on CLI."""

    def __call__(self, parser, namespace, values, option_string=None) -> None:
        """With can user can input dicts on CLI."""
        setattr(namespace, self.dest, dict())
        for value in values:
            # maxsplit helps in case righthand side of input has = in it, like filenames can have
            key, value = value.split("=", maxsplit=1)
            # catch list case
            if value.startswith("[") and value.endswith("]"):
                # if "[" in value and "]" in value:
                value = value.strip("][").split(",")
            # change numbers to numbers but with attention to decimals and negative numbers
            if is_int(value):
                value = int(value)
            elif is_float(value):
                value = float(value)
            elif is_None(value):
                value = None
            elif is_deltastr(value):
                value = pd.Timedelta(value).isoformat()
            elif is_datestr(value):
                value = pd.Timestamp(value)
            getattr(namespace, self.dest)[key] = value


def main():
    """Parser method.

    Include all inputs

    Example
    -------

    >>> python cli.py lon=-151 lat=59 use_auto_landmask=True start_time=2000-1-1
    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "kwargs",
        nargs="*",
        action=ParseKwargs,
        help="Input keyword arguments for running PTM. Available options are specific to the `catalog_type`. Dictionary-style input, e.g. `case='Oil'`. Format for list items is e.g. standard_names='[sea_water_practical_salinity,sea_water_temperature]'.",
    )

    parser.add_argument(
        "--dry-run",
        help="Return configuration parameters without running the model.",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    args = parser.parse_args()

    to_bool = {
        key: ast.literal_eval(value)
        for key, value in args.kwargs.items()
        if value in ["True", "False"]
    }
    args.kwargs.update(to_bool)

    if "output_file" not in args.kwargs:

        args.kwargs[
            "output_file"
        ] = f"output-results_{datetime.utcnow():%Y-%m-%dT%H%M:%SZ}.nc"

    # this is for running plots at the same time as a simulation
    # Convert the string representation of the dictionary to an actual dictionary
    if "plots" in args.kwargs:
        args.kwargs["plots"] = ast.literal_eval(args.kwargs["plots"])
    else:
        args.kwargs["plots"] = None

    # this is for running plots for an existing simulation file.
    # catch plots first
    if (args.kwargs["plots"] is not None) and not (
        set(args.kwargs) - set(["plots", "output_file"])
    ):

        import particle_tracking_manager.models.opendrift.plot as plot

        plots_dict = plot.make_plots_after_simulation(
            args.kwargs["output_file"], plots=args.kwargs["plots"]
        )
        print(plots_dict)
        return

    m = ptm.OpenDriftModel(**args.kwargs)  # , plots=plots)

    if args.dry_run:

        # run this to make sure everything is updated fully
        m.add_reader()
        print(m.config.model_dump())

    else:

        m.add_reader()

        m.seed()
        m.run()

        print(m.files.output_file)


if __name__ == "__main__":
    main()



"""Defines ParticleTrackingState and SetupOutputFiles."""

# Standard library imports
import datetime
import logging
import pathlib

from os import PathLike
from typing import Any, Mapping, Self

# Third-party imports
from pydantic import BaseModel, Field, computed_field, field_validator, model_validator

# Local imports
from .config_the_manager import OutputFormatEnum, TheManagerConfig


OUTPUT_FORMAT_SUFFIX: Mapping[OutputFormatEnum, str] = {
    OutputFormatEnum.netcdf: ".nc",
    OutputFormatEnum.parquet: ".parquet",  # can't be read back in by OpenDrift which is necessary for plots in PTM
    OutputFormatEnum.both: ".nc",  # first netcdf, then additionally output parquet at the end
}

logger = logging.getLogger()


class ParticleTrackingState(BaseModel):
    """Track simulation state."""

    has_run_setup: bool = False  # this may not be required for all models
    has_added_reader: bool = False
    has_run_seeding: bool = False
    has_run: bool = False


def generate_default_output_file() -> str:
    """Generate a default output file name based on the current date and time."""
    return f"output-results_{datetime.datetime.now():%Y-%m-%dT%H%M%SZ}"


class SetupOutputFiles(BaseModel):
    """Handle all changes/work on output files.

    This class runs first thing. Then logger setup.
    """

    output_file: PathLike[str] | None = Field(
        TheManagerConfig.model_json_schema()["properties"]["output_file"]["default"]
    )
    output_format: OutputFormatEnum = Field(
        TheManagerConfig.model_json_schema()["properties"]["output_format"]["default"]
    )

    model_config = {"validate_default": True}

    @field_validator("output_file", mode="after")
    def assign_output_file_if_needed(value: Any) -> str:
        """Assign a default output file name if not provided."""
        if value is None:
            return generate_default_output_file()
        return value

    @model_validator(mode="after")
    def set_output_file_extension(self) -> Self:
        """Set the appropriate file extension based on the output format."""
        assert self.output_file is not None
        if self.output_format not in OUTPUT_FORMAT_SUFFIX:
            raise ValueError(f"output_format {self.output_format} not recognized.")

        self.output_file = pathlib.Path(self.output_file).with_suffix(
            OUTPUT_FORMAT_SUFFIX[self.output_format]
        )

        return self

    @computed_field
    def logfile_name(self) -> str:
        """Generate a log file name based on the output file name."""
        assert self.output_file is not None
        return str(pathlib.Path(self.output_file).with_suffix(".log"))



"""Defines classes OpenDriftConfig, LeewayModelConfig, OceanDriftModelConfig, OpenOilModelConfig, LarvalFishModelConfig, HarmfulAlgalBloomModelConfig."""

# Standard library imports
import logging

from os import PathLike
from pathlib import Path
from typing import Any

from pydantic import Field, model_validator
from pydantic.fields import FieldInfo
from typing_extensions import Self

# Third-party imports
from particle_tracking_manager.ocean_model_registry import get_model_end_time

from ...config_ocean_model import register_on_the_fly

# Local imports
from ...config_the_manager import TheManagerConfig
from ...ocean_model_registry import get_model_end_time
from .enums import (
    CoastlineActionEnum,
    DiffusivityModelEnum,
    DriftModelEnum,
    DropletSizeDistributionEnum,
    HABSpeciesTypeEnum,
    ObjectTypeEnum,
    OilTypeEnum,
    PlotTypeEnum,
    RadiusTypeEnum,
    SeafloorActionEnum,
)
from .enums.species_types import (
    HAB_SPECIES_LABELS,
    SPECIES_HAB_DEFAULTS,
    HABParameters,
    _species_descriptions,
)


logger = logging.getLogger()

# class OpenDriftConfig(BaseModel):
class OpenDriftConfig(TheManagerConfig):
    """Some of the parameters in this mirror OpenDriftSimulation clss in OpenDrift"""

    drift_model: DriftModelEnum = Field(
        default=DriftModelEnum.OceanDrift,  # .value,
        description="Scenario to use for simulation.",
    )

    save_interpolator: bool = Field(
        default=False, description="Whether to save the interpolator."
    )

    interpolator_filename: PathLike[str] | None = Field(
        None,
        description="Filename to save interpolator to or read interpolator from. Exclude suffix (which should be .pickle).",
        json_schema_extra=dict(ptm_level=3),
    )

    export_variables: list[str] | None = Field(
        default=None,
        description="List of variables to export. Options available with `m.all_export_variables` for a given `drift_model`. "
        "['lon', 'lat', 'ID', 'status', 'z'] will always be exported. Default of None means all possible variables are exported.",
        json_schema_extra=dict(ptm_level=3),
    )

    plots: dict[str, dict] | str | None = Field(
        default=None,
        json_schema_extra=dict(ptm_level=1),
        description="Dictionary of plots to generate using OpenDrift.",
    )

    radius: float = Field(
        default=1000.0,
        ge=0.0,
        le=1000000,
        description="Radius around each lon-lat pair, within which particles will be seeded according to `radius_type`.",
        json_schema_extra=dict(
            ptm_level=2,
            units="m",
        ),
    )

    radius_type: RadiusTypeEnum = Field(
        default=RadiusTypeEnum.gaussian,  # .value,
        description="Distribution for seeding particles around location. Options: 'gaussian' or 'uniform'.",
        json_schema_extra=dict(
            ptm_level=3,
        ),
    )

    # OpenDriftSimulation parameters

    max_speed: float = Field(
        default=20.0,
        description="Typical maximum speed of elements, used to estimate reader buffer size",
        gt=0,
        title="Maximum speed",
        json_schema_extra={
            "units": "m/s",
            "od_mapping": "drift:max_speed",
            "ptm_level": 1,
        },
    )

    use_auto_landmask: bool = Field(
        default=False,
        description="If True, use a global-scale land mask from https://www.generic-mapping-tools.org/remote-datasets/earth-mask.html. Dataset scale selected is `auto`. If False, use the land mask from the ocean model.",
        title="Use Auto Landmask",
        json_schema_extra={"od_mapping": "general:use_auto_landmask", "ptm_level": 3},
    )

    coastline_action: CoastlineActionEnum = Field(
        default=CoastlineActionEnum.stranding,  # .value,
        description="This controls particle behavior at the coastline. Use `previous` for a particle to move back to its previous location if it hits land. Use `stranding` to have a particle stick (that is, become deactivated) where it interacts with land. With None, objects may also move over land.",
        title="Coastline Action",
        json_schema_extra={"od_mapping": "general:coastline_action", "ptm_level": 2},
    )

    current_uncertainty: float = Field(
        default=0,
        description="Add gaussian perturbation with this standard deviation to current components at each time step",
        title="Current Uncertainty",
        ge=0,
        le=5,
        json_schema_extra={
            "units": "m/s",
            "od_mapping": "drift:current_uncertainty",
            "ptm_level": 2,
        },
    )

    wind_uncertainty: float = Field(
        default=0,
        description="Add gaussian perturbation with this standard deviation to wind components at each time step.",
        title="Wind Uncertainty",
        ge=0,
        le=5,
        json_schema_extra={
            "units": "m/s",
            "od_mapping": "drift:wind_uncertainty",
            "ptm_level": 2,
        },
    )

    # add od_mapping to what should otherwise be in TheManagerConfig
    horizontal_diffusivity: float | None = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["horizontal_diffusivity"],
        Field(json_schema_extra=dict(od_mapping="drift:horizontal_diffusivity")),
    )
    stokes_drift: bool = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["stokes_drift"],
        Field(json_schema_extra=dict(od_mapping="drift:stokes_drift")),
    )
    z: float = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["z"],
        Field(json_schema_extra=dict(od_mapping="seed:z")),
    )
    number: int = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["number"],
        Field(json_schema_extra=dict(od_mapping="seed:number")),
    )

    model_config = {
        "validate_defaults": True,
        "use_enum_values": True,
        "extra": "forbid",
    }

    @model_validator(mode="after")
    def check_interpolator_filename(self) -> Self:
        """Check if interpolator_filename is set correctly."""
        if self.interpolator_filename is not None and not self.use_cache:
            raise ValueError(
                "If interpolator_filename is input, use_cache must be True."
            )
        return self

    # @model_validator(mode="after")
    # def check_config_z_value(self) -> Self:
    #     """Check if z is set correctly."""
    #     if hasattr(self, "seed_seafloor"):
    #         if not self.seed_seafloor and self.z is None:
    #             raise ValueError("z needs a non-None value if seed_seafloor is False.")
    #         if self.seed_seafloor and self.z is not None:
    #             raise ValueError("z needs to be None if seed_seafloor is True.")
    #     return self

    # this is not true! For example, OpenOil has by default no vertical advection but yes vertical mixing
    # @model_validator(mode="after")
    # def check_config_do3D(self) -> Self:
    #     """Check if do3D is set correctly."""
    #     if hasattr(self, "vertical_mixing"):
    #         if not self.do3D and self.vertical_mixing:
    #             raise ValueError(
    #                 "If do3D is False, vertical_mixing must also be False."
    #             )
    #     return self

    @model_validator(mode="after")
    def setup_interpolator(self) -> Self:
        """Setup interpolator."""

        if self.use_cache:
            if self.interpolator_filename is None:
                import appdirs

                cache_dir = Path(
                    appdirs.user_cache_dir(
                        appname="particle-tracking-manager",
                        appauthor="axiom-data-science",
                    )
                )
                cache_dir.mkdir(parents=True, exist_ok=True)
                self.interpolator_filename = cache_dir / Path(
                    f"{self.ocean_model}_interpolator"
                ).with_suffix(".pickle")
            else:
                self.interpolator_filename = Path(
                    self.interpolator_filename
                ).with_suffix(".pickle")
            self.save_interpolator = True

            # # change interpolator_filename to string
            # self.interpolator_filename = str(self.interpolator_filename)

            logger.debug(f"Interpolator filename: {self.interpolator_filename}")

            if Path(self.interpolator_filename).exists():
                logger.debug(
                    f"Will load the interpolator from {self.interpolator_filename}."
                )
            else:
                logger.debug(
                    f"A new interpolator will be saved to {self.interpolator_filename}."
                )

        else:
            self.save_interpolator = False
            logger.debug("Interpolator will not be saved.")

        return self

    @property
    def drop_vars(self) -> list[str]:
        """Gather variables to drop based on PTMConfig and OpenDriftConfig."""

        # set drop_vars initial values based on the PTM settings, then add to them for the specific model
        drop_vars = (
            self.ocean_model_config.model_drop_vars.copy()
        )  # without copy this remembers drop_vars from other instances

        # don't need w if not 3D movement
        if not self.do3D:
            drop_vars += ["w"]
            logger.debug("Dropping vertical velocity (w) because do3D is False")
        else:
            logger.debug("Retaining vertical velocity (w) because do3D is True")

        # don't need winds if stokes drift, wind drift, added wind_uncertainty, and vertical_mixing are off
        # It's possible that winds aren't required for every OpenOil simulation but seems like
        # they would usually be required and the cases are tricky to navigate so also skipping for that case.
        if (
            not self.stokes_drift
            and (hasattr(self, "wind_drift_factor") and self.wind_drift_factor == 0)
            and self.wind_uncertainty == 0
            and self.drift_model != "OpenOil"
            and not self.vertical_mixing
        ):
            drop_vars += ["Uwind", "Vwind", "Uwind_eastward", "Vwind_northward"]
            logger.debug(
                "Dropping wind variables because stokes_drift, wind_drift_factor, wind_uncertainty, and vertical_mixing are all off and drift_model is not 'OpenOil'"
            )
        else:
            logger.debug(
                "Retaining wind variables because stokes_drift, wind_drift_factor, wind_uncertainty, or vertical_mixing are on or drift_model is 'OpenOil'"
            )

        # only keep salt and temp for LarvalFish or OpenOil
        if self.drift_model not in ["LarvalFish", "OpenOil", "HarmfulAlgalBloom"]:
            drop_vars += ["salt", "temp"]
            logger.debug(
                "Dropping salt and temp variables because drift_model is not LarvalFish nor OpenOil nor HarmfulAlgalBloom"
            )
        else:
            logger.debug(
                "Retaining salt and temp variables because drift_model is LarvalFish or OpenOil or HarmfulAlgalBloom"
            )

        # keep some ice variables for OpenOil (though later see if these are used)
        if self.drift_model != "OpenOil":
            drop_vars += ["aice", "uice_eastward", "vice_northward"]
            logger.debug("Dropping ice variables because drift_model is not OpenOil")
        else:
            logger.debug("Retaining ice variables because drift_model is OpenOil")

        # if using static masks, drop wetdry masks.
        # if using wetdry masks, drop static masks.
        # TODO: is standard_name_mapping working correctly?
        if self.use_static_masks:
            # TODO: Can the mapping include all possible mappings or does it need to be exact?
            # standard_name_mapping.update({"mask_rho": "land_binary_mask"})
            drop_vars += ["wetdry_mask_rho", "wetdry_mask_u", "wetdry_mask_v"]
            logger.debug("Dropping wetdry masks because using static masks instead.")
        else:
            # standard_name_mapping.update({"wetdry_mask_rho": "land_binary_mask"})
            drop_vars += ["mask_rho", "mask_u", "mask_v", "mask_psi"]
            logger.debug(
                "Dropping mask_rho, mask_u, mask_v, mask_psi because using wetdry masks instead."
            )
        return drop_vars

    @model_validator(mode="after")
    def check_plot_oil(self) -> Self:
        """Check if oil budget plot is requested and drift model is OpenOil."""
        if (
            self.plots is not None
            and isinstance(self.plots, dict)
            and "oil" in self.plots.keys()
        ):
            if self.drift_model != "OpenOil":
                raise ValueError(
                    "Oil budget plot only available for OpenOil drift model"
                )
        return self

    @model_validator(mode="after")
    def check_plot_all(self) -> Self:
        """Check if all plots are requested."""
        if (
            self.plots is not None
            and isinstance(self.plots, dict)
            and "all" in self.plots.keys()
            and len(self.plots) > 1
        ):
            raise ValueError(
                "If 'all' is specified for plots, it must be the only plot option."
            )
        return self

    @model_validator(mode="after")
    def check_plot_prefix_enum(self) -> Self:
        """Check if plot keys start with a PlotTypeEnum."""
        if self.plots is not None:
            assert isinstance(self.plots, dict)
            present_keys = [
                key
                for key in self.plots.keys()
                for PlotType in PlotTypeEnum
                if key.startswith(PlotType.value)
            ]
            random_keys = set(self.plots.keys()) - set(present_keys)
            if len(random_keys) > 0:
                raise ValueError(
                    f"Plot keys must start with a PlotTypeEnum. The following keys do not: {random_keys}"
                )
        return self


class LeewayModelConfig(OpenDriftConfig):
    """Leeway model configuration for OpenDrift."""

    drift_model: DriftModelEnum = DriftModelEnum.Leeway

    object_type: ObjectTypeEnum = Field(
        default=ObjectTypeEnum("Person-in-water (PIW), unknown state (mean values)"),
        description="Leeway object category for this simulation. List is originally from USCG technical reports. More details here: https://github.com/OpenDrift/opendrift/blob/master/opendrift/models/OBJECTPROP.DAT.",
        title="Object Type",
        json_schema_extra={"od_mapping": "seed:object_type", "ptm_level": 1},
    )

    # modify default values
    stokes_drift: bool = FieldInfo.merge_field_infos(
        OpenDriftConfig.model_fields["stokes_drift"], Field(default=False)
    )

    @model_validator(mode="after")
    def check_stokes_drift(self) -> Self:
        """Check if stokes_drift is set to False for Leeway model."""
        if self.stokes_drift:
            raise ValueError("stokes_drift must be False with the Leeway drift model.")

        return self

    @model_validator(mode="after")
    def check_do3D(self) -> Self:
        """Check if do3D is set to False for Leeway model."""
        if self.do3D:
            raise ValueError("do3D must be False with the Leeway drift model.")

        return self


class OceanDriftModelConfig(OpenDriftConfig):
    """Ocean drift model configuration for OpenDrift."""

    drift_model: DriftModelEnum = DriftModelEnum.OceanDrift  # .value

    seed_seafloor: bool = Field(
        default=False,
        description="Elements are seeded at seafloor, and seeding depth (z) is neglected and must be None.",
        title="Seed Seafloor",
        json_schema_extra={"od_mapping": "seed:seafloor", "ptm_level": 2},
    )

    diffusivitymodel: DiffusivityModelEnum = Field(
        default="windspeed_Large1994",
        description="Algorithm/source used for profile of vertical diffusivity. Environment means that diffusivity is acquired from readers or environment constants/fallback. Parameterizations based on wind speed are also available.",
        title="Diffusivity model",
        json_schema_extra={
            "units": "seconds",
            "od_mapping": "vertical_mixing:diffusivitymodel",
            "ptm_level": 3,
        },
    )

    mixed_layer_depth: float = Field(
        default=20,
        description="mixed_layer_depth controls how deep the vertical diffusivity profile reaches. This sets the fallback value for ocean_mixed_layer_thickness if not available from any reader.",
        title="Mixed Layer Depth",
        ge=0.0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "environment:constant:ocean_mixed_layer_thickness",
            "ptm_level": 3,
        },
    )

    seafloor_action: SeafloorActionEnum = Field(
        default=SeafloorActionEnum.lift_to_seafloor,  # .value,
        description="This controls particle behavior at the seafloor. Use `deactivate` to stick particles to the seafloor at the point of interaction. Use `lift_to_seafloor` to elevate particles up to seabed if below. User `previous` to move elements back to previous position. Use None to ignore seafloor.",
        title="Seafloor Action",
        json_schema_extra={
            "od_mapping": "general:seafloor_action",
            "ptm_level": 2,
        },
    )

    wind_drift_depth: float = Field(
        default=0.1,
        description="The direct wind drift (windage) is linearly decreasing from the surface value (wind_drift_factor) until 0 at this depth.",
        title="Wind Drift Depth",
        ge=0,
        le=1,
        json_schema_extra={
            "units": "meters",
            "od_mapping": "drift:wind_drift_depth",
            "ptm_level": 3,
        },
    )

    vertical_mixing_timestep: float = Field(
        default=60,
        description="Time step used for inner (fast) loop of the vertical mixing model. Set this smaller to increase frequency of vertical mixing calculation; number of loops is calculated as int(self.time_step/vertical_mixing_timestep) so vertical_mixing_timestep must be smaller than time_step.",
        title="Vertical Mixing Timestep",
        ge=0.1,
        le=3600,
        json_schema_extra={
            "units": "seconds",
            "od_mapping": "vertical_mixing:timestep",
            "ptm_level": 3,
        },
    )

    wind_drift_factor: float = Field(
        default=0.02,
        description="Elements at surface are moved with this fraction of the wind vector, in addition to currents and Stokes drift. Multiply by 100 to get the percent windage.",
        title="Wind Drift Factor",
        ge=0,
        le=0.1,
        json_schema_extra={
            "units": "1",
            "od_mapping": "seed:wind_drift_factor",
            "ptm_level": 2,
        },
    )

    vertical_mixing: bool = Field(
        default=False,
        description="Activate vertical mixing scheme. Vertical mixing includes movement due to buoyancy and turbulent mixing.",
        title="Vertical Mixing",
        json_schema_extra={
            "od_mapping": "drift:vertical_mixing",
            "ptm_level": 2,
        },
    )

    vertical_mixing_at_surface: bool = Field(
        default=True,
        description="If vertical mixing is activated, surface elements (z=0) can only be mixed (downwards) if this setting it True.",
        title="Vertical Mixing At Surface",
        json_schema_extra={
            "od_mapping": "drift:vertical_mixing_at_surface",
            "ptm_level": 2,
        },
    )

    vertical_advection_at_surface: bool = Field(
        default=True,
        description="If vertical advection is activated, surface elements (z=0) can only be advected (downwards) if this setting it True.",
        title="Vertical Advection At Surface",
        json_schema_extra={
            "od_mapping": "drift:vertical_advection_at_surface",
            "ptm_level": 2,
        },
    )


class OpenOilModelConfig(OceanDriftModelConfig):
    """OpenOil model configuration for OpenDrift."""

    drift_model: DriftModelEnum = DriftModelEnum.OpenOil  # .value

    oil_type: OilTypeEnum = Field(
        default=OilTypeEnum.AD04012,  # .value,
        description="Oil type to be used for the simulation, from the NOAA ADIOS database.",
        title="Oil Type",
        json_schema_extra={
            "od_mapping": "seed:oil_type",
            "ptm_level": 1,
            "oneOf": [{"const": oil.value, "title": oil.title} for oil in OilTypeEnum],
        },
    )

    m3_per_hour: float = Field(
        default=1,
        description="The amount (volume) of oil released per hour (or total amount if release is instantaneous).",
        title="M3 Per Hour",
        gt=0,
        json_schema_extra={
            "units": "m3 per hour",
            "od_mapping": "seed:m3_per_hour",
            "ptm_level": 1,
        },
    )

    oil_film_thickness: float = Field(
        default=0.001,
        description="Seeding value of oil_film_thickness. Values are calculated by OpenDrift starting from this initial value if `update_oilfilm_thickness==True`.",
        title="Oil Film Thickness",
        json_schema_extra={
            "units": "m",
            "od_mapping": "seed:oil_film_thickness",
            "ptm_level": 3,
        },
    )

    droplet_size_distribution: DropletSizeDistributionEnum = Field(
        default=DropletSizeDistributionEnum.uniform,  # .value,
        description="Droplet size distribution used for subsea release.",
        title="Droplet Size Distribution",
        json_schema_extra={
            "od_mapping": "seed:droplet_size_distribution",
            "ptm_level": 3,
        },
    )

    droplet_diameter_mu: float = Field(
        default=0.001,
        description="The mean diameter of oil droplet for a subsea release, used in normal/lognormal distributions.",
        title="Droplet Diameter Mu",
        ge=1e-08,
        le=1,
        json_schema_extra={
            "units": "meters",
            "od_mapping": "seed:droplet_diameter_mu",
            "ptm_level": 3,
        },
    )

    droplet_diameter_sigma: float = Field(
        default=0.0005,
        description="The standard deviation in diameter of oil droplet for a subsea release, used in normal/lognormal distributions.",
        title="Droplet Diameter Sigma",
        ge=1e-08,
        le=1,
        json_schema_extra={
            "units": "meters",
            "od_mapping": "seed:droplet_diameter_sigma",
            "ptm_level": 3,
        },
    )

    droplet_diameter_min_subsea: float = Field(
        default=0.0005,
        description="The minimum diameter of oil droplet for a subsea release, used in uniform distribution.",
        title="Droplet Diameter Min Subsea",
        ge=1e-08,
        le=1,
        json_schema_extra={
            "units": "meters",
            "od_mapping": "seed:droplet_diameter_min_subsea",
            "ptm_level": 3,
        },
    )

    droplet_diameter_max_subsea: float = Field(
        default=0.005,
        description="The maximum diameter of oil droplet for a subsea release, used in uniform distribution.",
        title="Droplet Diameter Max Subsea",
        ge=1e-08,
        le=1,
        json_schema_extra={
            "units": "meters",
            "od_mapping": "seed:droplet_diameter_max_subsea",
            "ptm_level": 3,
        },
    )

    emulsification: bool = Field(
        default=True,
        description="If True, surface oil is emulsified, i.e. water droplets are mixed into oil due to wave mixing, with resulting increase of viscosity.",
        title="Emulsification",
        json_schema_extra={
            "od_mapping": "processes:emulsification",
            "ptm_level": 2,
        },
    )

    dispersion: bool = Field(
        default=True,
        description="If True, oil is removed from simulation (dispersed), if entrained as very small droplets.",
        title="Dispersion",
        json_schema_extra={
            "od_mapping": "processes:dispersion",
            "ptm_level": 2,
        },
    )

    evaporation: bool = Field(
        default=True,
        description="If True, surface oil is evaporated.",
        title="Evaporation",
        json_schema_extra={
            "od_mapping": "processes:evaporation",
            "ptm_level": 2,
        },
    )

    update_oilfilm_thickness: bool = Field(
        default=True,
        description="If True, Oil film thickness is calculated at each time step. If False, oil film thickness is kept constant with value provided at seeding.",
        title="Update Oilfilm Thickness",
        json_schema_extra={
            "od_mapping": "processes:update_oilfilm_thickness",
            "ptm_level": 2,
        },
    )

    biodegradation: bool = Field(
        default=True,
        description="If True, oil mass is biodegraded (eaten by bacteria).",
        title="Biodegradation",
        json_schema_extra={
            "od_mapping": "processes:biodegradation",
            "ptm_level": 2,
        },
    )

    # overwrite the defaults from OceanDriftModelConfig for a few inherited parameters,
    # but don't want to have to repeat the full definition
    current_uncertainty: float = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["current_uncertainty"], Field(default=0.0)
    )
    wind_uncertainty: float = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["wind_uncertainty"], Field(default=0.0)
    )
    wind_drift_factor: float = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["wind_drift_factor"], Field(default=0.03)
    )
    vertical_mixing: bool = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["vertical_mixing"], Field(default=True)
    )
    vertical_mixing_at_surface: bool = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["vertical_mixing_at_surface"],
        Field(default=False),
    )
    vertical_advection_at_surface: bool = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["vertical_advection_at_surface"],
        Field(default=False),
    )


class LarvalFishModelConfig(OceanDriftModelConfig):
    """Larval fish model configuration for OpenDrift."""

    drift_model: DriftModelEnum = DriftModelEnum.LarvalFish  # .value

    diameter: float = Field(
        default=0.0014,
        description="Seeding value of diameter. The diameter gives the egg diameter so must be used with `hatched=0`.",
        title="Diameter",
        gt=0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "seed:diameter",
            "ptm_level": 2,
        },
    )

    neutral_buoyancy_salinity: float = Field(
        default=31.25,
        description="Seeding value of neutral_buoyancy_salinity. This is a property of the egg so must be used with `hatched=0`.",
        title="Neutral Buoyancy Salinity",
        gt=0,
        json_schema_extra={
            "units": "PSU",
            "od_mapping": "seed:neutral_buoyancy_salinity",
            "ptm_level": 2,
        },
    )

    stage_fraction: float = Field(
        default=0.0,
        description="Seeding value of stage_fraction. stage_fraction tracks percentage of development time completed, from 0 to 1, where a value of 1 means the egg has hatched. If `hatched==1` then `stage_fraction` is ignored in `OpenDrift`.",
        title="Stage Fraction",
        ge=0,
        le=1,
        json_schema_extra={
            "units": "",
            "od_mapping": "seed:stage_fraction",
            "ptm_level": 2,
        },
    )

    hatched: int = Field(
        default=0,
        description="Seeding value of hatched. 0 for eggs, 1 for larvae.",
        title="Hatched",
        ge=0,
        le=1,
        json_schema_extra={
            "units": "",
            "od_mapping": "seed:hatched",
            "ptm_level": 2,
        },
    )

    length: float | None = Field(
        default=None,
        description="Seeding value of length. This is not currently used.",
        title="Length",
        gt=0,
        json_schema_extra={
            "units": "mm",
            "od_mapping": "seed:length",
            "ptm_level": 2,
        },
    )

    weight: float = Field(
        default=0.08,
        description="Seeding value of weight. This is the starting weight for larval fish, whenever they reach that stage.",
        title="Weight",
        gt=0,
        json_schema_extra={
            "units": "mg",
            "od_mapping": "seed:weight",
            "ptm_level": 2,
        },
    )

    # override inherited parameter defaults
    vertical_mixing: bool = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["vertical_mixing"], Field(default=True)
    )
    do3D: bool = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["do3D"], Field(default=True)
    )

    @model_validator(mode="after")
    def check_do3D(self) -> Self:
        """Check if do3D is set to True for LarvalFish model."""
        if not self.do3D:
            raise ValueError("do3D must be True with the LarvalFish drift model.")

        return self

    @model_validator(mode="after")
    def check_vertical_mixing(self) -> Self:
        """Check if vertical_mixing is set to True for LarvalFish model."""
        if not self.vertical_mixing:
            raise ValueError(
                "vertical_mixing must be True with the LarvalFish drift model."
            )

        return self

    # @model_validator(mode="after")
    # def check_hatched_and_stage_fraction(self) -> Self:
    #     """If hatched==1, stage_fraction should be None.

    #     This only applies for seeding, not for the simulation.
    #     """

    #     if self.hatched == 1 and self.stage_fraction is not None:
    #         raise ValueError("If hatched==1, stage_fraction should be None.")
    #     return self


class HarmfulAlgalBloomModelConfig(HABParameters, OceanDriftModelConfig):
    """Harmful algal bloom model configuration for OpenDrift."""

    drift_model: DriftModelEnum = DriftModelEnum.HarmfulAlgalBloom

    species_type: HABSpeciesTypeEnum = Field(
        default=HABSpeciesTypeEnum.PN,
        description="HarmfulAlgalBloom species category for this simulation. This option maps to individual properties which can instead be set manually if desired.",
        title="HAB Species Type",
        json_schema_extra={
            "ptm_level": 1,
            "oneOf": [
                {
                    "const": species.value,
                    "title": HAB_SPECIES_LABELS[species],
                    "description": _species_descriptions[species.value],
                }
                for species in HABSpeciesTypeEnum
            ],
        },
    )

    # override inherited parameter defaults
    vertical_mixing: bool = FieldInfo.merge_field_infos(
        OceanDriftModelConfig.model_fields["vertical_mixing"], Field(default=True)
    )
    do3D: bool = FieldInfo.merge_field_infos(
        TheManagerConfig.model_fields["do3D"], Field(default=True)
    )

    @model_validator(mode="before")
    @classmethod
    def apply_species_defaults(cls, data: Any) -> Any:
        """
        - If species_type has presets:
            * Start from species HAB defaults
            * Overlay any user-provided hab_params
            * Set z/do3D defaults only if user didn't supply them
        - If species_type has no presets (e.g., Custom):
            * Require user to provide hab_params explicitly
            * z/do3D behave as usual (user or base defaults)
        """
        if not isinstance(data, dict):
            return data

        # Ensure species_type has some value in raw input
        data.setdefault("species_type", HABSpeciesTypeEnum.PN)
        species = data["species_type"]

        # -------- HAB param defaults (flattened) --------
        if species in SPECIES_HAB_DEFAULTS:
            # species defaults
            default_params = SPECIES_HAB_DEFAULTS[species].model_dump()

            # apply defaults only where user did not provide a value
            for field_name, default_value in default_params.items():
                data.setdefault(field_name, default_value)
        else:
            # species has no preset (e.g. Custom): require all HAB fields
            required_fields = HABParameters.model_fields.keys()
            missing = [f for f in required_fields if f not in data]
            if missing:
                raise ValueError(
                    "Custom HAB species requires explicit values for all HAB parameters. "
                    f"Missing: {', '.join(missing)}"
                )

        # # -------- Other config field defaults (e.g. z, do3D) --------
        # field_defaults = SPECIES_HAB_MANAGER_DEFAULTS.get(species, {})
        # for field_name, default_value in field_defaults.items():
        #     # only apply if the user did not supply this field
        #     data.setdefault(field_name, default_value)

        return data

    @model_validator(mode="after")
    def setup_species_parameters(self) -> Self:
        """Assign species-specific parameters that depend on other inputs."""

        MLD = self.mixed_layer_depth

        if self.species_type == HABSpeciesTypeEnum.PN:
            # Pseudo-nitzschia

            # want to cap the resulting max depth to 25m
            # a straight forward way to do this is to control the max MLD in this calculation
            effective_MLD = min(MLD, 31.25)  # 0.4*31.25 + 0.4*31.25 = 25

            # use the mixed layer depth to calculate band_center_depth and band_half_width
            self.band_center_depth = -0.4 * effective_MLD
            self.band_half_width = 0.4 * effective_MLD

        elif self.species_type == HABSpeciesTypeEnum.AX:
            # Alexandrium

            # want to cap the resulting diel_night_depth to about 40m
            # a straight forward way to do this is to control the max MLD in this calculation
            effective_MLD = min(MLD, 30.75)  # 1.3*30.75 = 39.975

            # use the mixed layer depth to calculate band_center_depth and band_half_width
            self.diel_day_depth = -0.3 * effective_MLD
            self.diel_night_depth = -1.3 * effective_MLD

        elif self.species_type == HABSpeciesTypeEnum.DP:
            # Dinophysis

            d = 2.5
            # want min depth to be about 5m
            # a straight forward way to do this is to control the max MLD in this calculation
            effective_MLD = max(MLD, 5)  # -(MLD+d)+0.5*MLD with MLD=5

            # want to cap the resulting max depth to about 25m
            # a straight forward way to do this is to control the min MLD in this calculation
            effective_MLD = min(effective_MLD, 15)  # -(MLD+d)-0.5*MLD with MLD=15

            # use the mixed layer depth to calculate band_center_depth and band_half_width
            self.band_center_depth = -(effective_MLD + d)
            self.band_half_width = 0.5 * effective_MLD

        else:
            raise ValueError(
                f"Species type {self.species_type} not recognized for HarmfulAlgalBloom model."
            )
        return self

    #     if self.species_type == HABSpeciesTypeEnum("Pseudo_nitzschia"):

    #         if self.do3D:
    #             raise ValueError("Pseudo_nitzschia species requires do3D to be False.")
    #         if self.z != 0.0:
    #             raise ValueError("Pseudo_nitzschia species requires z to be 0.0 m.")
    #         logger.debug("HAB species Pseudo_nitzschia selected.")

    #     else:
    #         raise ValueError(
    #             f"Species type {self.species_type} not recognized for HarmfulAlgalBloom model."
    #         )
    #     return self

    # @model_validator(mode="after")
    # def check_do3D(self) -> Self:
    #     """Check if do3D is set to True for LarvalFish model."""
    #     if not self.do3D:
    #         raise ValueError("do3D must be True with the LarvalFish drift model.")

    #     return self

    # @model_validator(mode="after")
    # def check_vertical_mixing(self) -> Self:
    #     """Check if vertical_mixing is set to True for LarvalFish model."""
    #     if not self.vertical_mixing:
    #         raise ValueError(
    #             "vertical_mixing must be True with the LarvalFish drift model."
    #         )

    #     return self


open_drift_mapper: dict[str, type[OpenDriftConfig]] = {
    "OceanDrift": OceanDriftModelConfig,
    "OpenOil": OpenOilModelConfig,
    "LarvalFish": LarvalFishModelConfig,
    "Leeway": LeewayModelConfig,
    "HarmfulAlgalBloom": HarmfulAlgalBloomModelConfig,
}



"""Test manager use in library, the default approach."""


from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest

from pydantic import ValidationError

from particle_tracking_manager.config_the_manager import TheManagerConfig
from particle_tracking_manager.the_manager import ParticleTrackingManager


class TestConfig(TheManagerConfig):
    pass


# Set up a subclass for testing. This is meant to be a simple version of the
# OpenDriftModel.
class TestParticleTrackingManager(ParticleTrackingManager):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.config = TestConfig(**kwargs)

    def _add_reader(self):
        pass

    def _seed(self):
        pass

    def _run(self):
        pass

    def _setup_for_simulation(self):
        pass


def test_order():
    """Have to configure before seeding."""

    m = TestParticleTrackingManager(steps=1, start_time="2022-01-01")
    with pytest.raises(ValueError):
        m.run()


def test_seed_order():
    m = TestParticleTrackingManager(steps=1, start_time="2022-01-01")
    m.state.has_added_reader = True
    m.seed()

    m = TestParticleTrackingManager(steps=1, start_time="2022-01-01")
    with pytest.raises(ValueError):
        m.seed()


def test_log_name():
    m = TestParticleTrackingManager(output_file="newtest", steps=1)
    assert m.files.logfile_name == "newtest.log"

    m = TestParticleTrackingManager(output_file="newtest.nc", steps=1)
    assert m.files.logfile_name == "newtest.log"

    m = TestParticleTrackingManager(output_file="newtest.parq", steps=1)
    assert m.files.logfile_name == "newtest.log"

    m = TestParticleTrackingManager(output_file="newtest.parquet", steps=1)
    assert m.files.logfile_name == "newtest.log"


def test_output_file():
    """make sure output file is parquet if output_format is parquet"""

    m = TestParticleTrackingManager(output_format="parquet", steps=1)
    assert m.files.output_file.suffix == ".parquet"

    m = TestParticleTrackingManager(output_format="netcdf", steps=1)
    assert m.files.output_file.suffix == ".nc"


def test_ocean_model_not_known():
    with pytest.raises(ValidationError):
        TestParticleTrackingManager(ocean_model="wrong_name", steps=1)



"""Defines OceanModelSimulation and ocean_model_simulation_mapper, a dict with key, value pairs of ocean model name and OceanModelSimulation class. Each OceanModelSimulation is built using information from an OceanModelConfig instance.

Functions needed to work with ocean model output are defined here. Also register on-the-fly instances of ocean models here.
"""

# Standard library imports
import logging

from datetime import datetime, timedelta
from enum import Enum
from typing import Annotated, Any, Callable

# Third-party imports
import xarray as xr

from pydantic import BaseModel, Field, create_model, model_validator
from typing_extensions import Self

# Local imports
from .models.opendrift.utils import make_ciofs_kerchunk, make_nwgoa_kerchunk
from .ocean_model_registry import OceanModelConfig, ocean_model_registry


logger = logging.getLogger()


# Define a function to generate an Enum from the registry
def generate_enum_from_registry() -> Enum:
    """Generate an Enum class from the ocean model registry."""
    # Dynamically create an Enum class using the registry data
    enum_name = "OceanModelEnum"
    enum_members = {name: name for name in ocean_model_registry.all()}
    return Enum(enum_name, enum_members)


# Generate a dynamic Enum using the registry data
# since user might add their own models.
OceanModelEnum = generate_enum_from_registry()


## Set up ocean model simulation configuration: depends on a tracking simulation. ##


class OceanModelSimulation(BaseModel):
    """Ocean model simulation configuration."""

    ocean_model_local: bool

    model_config = {
        "validate_defaults": True,
        "use_enum_values": True,
        "extra": "forbid",
    }

    @model_validator(mode="after")
    def check_config_oceanmodel_lon0_360(self) -> Self:
        """Check if the ocean model is using 0-360 longitude convention."""
        if self.ocean_model_config.oceanmodel_lon0_360:
            if self.lon is not None and self.lon < 0:
                if -180 < self.lon < 0:
                    orig_lon = self.lon
                    self.lon += 360
                    logger.debug(f"Shifting longitude from {orig_lon} to {self.lon}.")
        return self

    def open_dataset(self, drop_vars: list) -> xr.Dataset:
        """Open an xarray dataset"""
        # if local
        if self.ocean_model_local:

            name, kerchunk_func_str = (
                self.ocean_model_config.name,
                self.ocean_model_config.kerchunk_func_str,
            )
            start_time, end_time = self.start_time, self.end_time

            if loc_local(name, kerchunk_func_str, start_time, end_time) is None:
                raise ValueError(
                    "loc_local must be set if ocean_model_local is True, but loc_local is None."
                )
            else:
                # TODO: Make a way to input chunks selection (and maybe other xarray kwargs)
                ds = xr.open_dataset(
                    loc_local(name, kerchunk_func_str, start_time, end_time),
                    engine="kerchunk",
                    # chunks={},  # Looks like it is faster not to include this for kerchunk
                    drop_variables=drop_vars,
                    decode_times=False,
                )
                logger.debug(
                    f"Opened local dataset with start time {start_time} and end time {end_time} and number outputs {ds.ocean_time.size}."
                )

        # otherwise remote
        else:
            if self.ocean_model_config.loc_remote is None:
                raise ValueError(
                    "loc_remote must be set if ocean_model_local is False, but loc_remote is None."
                )
            else:
                if ".nc" in self.ocean_model_config.loc_remote:
                    ds = xr.open_dataset(
                        self.ocean_model_config.loc_remote,
                        chunks=self.ocean_model_config.chunks,
                        drop_variables=drop_vars,
                        decode_times=False,
                    )
                elif ".parquet" in self.ocean_model_config.loc_remote:
                    ds = xr.open_dataset(
                        self.ocean_model_config.loc_remote,
                        engine="kerchunk",
                        # chunks={},  # Looks like it is faster not to include this for kerchunk
                        drop_variables=drop_vars,
                        decode_times=False,
                    )
                else:
                    ds = xr.open_zarr(
                        self.ocean_model_config.loc_remote,
                        chunks=self.ocean_model_config.chunks,
                        drop_variables=drop_vars,
                        decode_times=False,
                    )

                logger.debug(
                    f"Opened remote dataset {self.ocean_model_config.loc_remote} with number outputs {ds.ocean_time.size}."
                )
        return ds


def create_ocean_model_simulation(
    ocean_model: OceanModelConfig,
) -> type[OceanModelSimulation]:
    """Create an ocean model simulation object."""
    ocean_model_name = ocean_model.name
    simulation_model = create_model(
        ocean_model_name,  # Model name
        __base__=OceanModelSimulation,
        lon=(
            float | None,
            Field(
                ...,
                ge=getattr(ocean_model, "lon_min"),
                le=getattr(ocean_model, "lon_max"),
                description="Longitude of the simulation within the model bounds.",
            ),
        ),
        lat=(
            float | None,
            Field(
                ...,
                ge=getattr(ocean_model, "lat_min"),
                le=getattr(ocean_model, "lat_max"),
                description="Latitude of the simulation within the model bounds.",
            ),
        ),
        start_time=(
            datetime,
            Field(
                ...,
                ge=getattr(ocean_model, "start_time_model"),
                le=getattr(ocean_model, "end_time_model"),
                description="Start time of the simulation.",
            ),
        ),
        end_time=(
            datetime,
            Field(
                ...,
                ge=getattr(ocean_model, "start_time_model"),
                le=getattr(ocean_model, "end_time_model"),
                description="End time of the simulation.",
            ),
        ),
        ocean_model_config=(OceanModelConfig, ocean_model),
    )
    return simulation_model


ocean_model_simulation_mapper = {}
for ocean_model in ocean_model_registry.all_models():
    ocean_model_simulation_mapper[ocean_model.name] = create_ocean_model_simulation(
        ocean_model
    )


MAKE_KERCHUNK_FUNCTIONS: dict[
    str, Callable[[datetime, datetime, str], dict[Any, Any]]
] = {
    "make_nwgoa_kerchunk": make_nwgoa_kerchunk,
    "make_ciofs_kerchunk": make_ciofs_kerchunk,
}


def loc_local(
    name: str, kerchunk_func_str: str, start_sim: datetime, end_sim: datetime
) -> dict:
    """This sets up a short kerchunk file for reading in just enough model output."""

    # back each start time back 1 day and end time forward 1 day to make sure enough output is available
    if start_sim < end_sim:
        start_time = start_sim - timedelta(days=1)
        end_time = end_sim + timedelta(days=1)
    else:
        start_time = start_sim + timedelta(days=1)
        end_time = end_sim - timedelta(days=1)

    return MAKE_KERCHUNK_FUNCTIONS[kerchunk_func_str](start_time, end_time, name)


def register_on_the_fly(ds_info: dict, ocean_model: str = "ONTHEFLY") -> None:
    """Update an ocean model in the registry and rebuild its Pydantic model.

    The default model to register is "ONTHEFLY", which is a placeholder for user-defined models.
    However, alternations could also be made to any existing model in the registry.

    ds_info can contain any of the OceanModelConfig fields.
    """

    # Update the ocean model template with user dataset information
    ocean_model_registry.update_model(ocean_model, ds_info)

    # Create the ocean model simulation object for the new ocean model
    simulation_cls = create_ocean_model_simulation(
        ocean_model_registry.get(ocean_model)
    )

    # Update the ocean model simulation mapper with the new ocean model simulation
    ocean_model_simulation_mapper.update({ocean_model: simulation_cls})

    logger.info(
        "Registered new ocean model or altered existing ocean model in the registry."
    )


def update_TXLA_with_download_location() -> None:
    """The user-defined TXLA model is missing the download location

    because it depends on the user's operating system and setup. Run this
    function to update the TXLA model with the location of the downloaded
    file.
    """

    import xroms

    url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
    ds_info = dict(loc_remote=url)
    register_on_the_fly(ds_info, ocean_model="TXLA")
    logger.debug("Updated TXLA model with download location.")



"""Test seeding functionality."""

from datetime import datetime

import numpy as np
import pytest

import particle_tracking_manager as ptm


@pytest.mark.slow
def test_seeding_from_geojson():
    """Seed from GeoJSON."""

    geo = {
        "type": "Feature",
        "properties": {},
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [-90, 28.7],
                    [-90, 28.8],
                    [-90.1, 28.8],
                    [-90.1, 28.7],
                    [-90, 28.7],
                ]
            ],
        },
    }
    m = ptm.OpenDriftModel(
        geojson=geo,
        radius_type="uniform",
        radius=1000,
        use_auto_landmask=True,
        number=2,
        steps=1,
        lon=None,
        lat=None,
        ocean_model="TXLA",
        ocean_model_local=False,
        start_time=datetime(2009, 11, 19, 12, 0),
    )
    m.add_reader()
    m.seed()

    expected_lon = [-90.06466, -90.06964]
    expected_lat = [28.726435, 28.731668]

    assert np.allclose(m.initial_drifters.lon, expected_lon)
    assert np.allclose(m.initial_drifters.lat, expected_lat)
    assert m._seed_kws["radius_type"] == "uniform"
    assert m._seed_kws["radius"] == 1000



"""Uses OpenDrift for particle tracking model."""

# Standard library imports
import json
import logging

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any

import pandas as pd
import xarray as xr

# Third-party imports
from opendrift.readers import reader_ROMS_native

from particle_tracking_manager.models.opendrift.enums.oil_types import OIL_ID_TO_NAME

from ...config_the_manager import OutputFormatEnum

# Local imports
from ...ocean_model_registry import ocean_model_registry
from ...the_manager import ParticleTrackingManager
from .config_opendrift import OpenDriftConfig, open_drift_mapper
from .plot import make_plots
from .utils import (
    apply_known_ocean_model_specific_changes,
    apply_user_input_ocean_model_specific_changes,
    narrow_dataset_to_simulation_time,
)


# Initialize logger
logger = logging.getLogger()


class OpenDriftModel(ParticleTrackingManager):
    """OpenDrift particle tracking model.

    Parameters
    ----------
    drift_model : str
        Options: "OceanDrift", "LarvalFish", "OpenOil", "Leeway", "HarmfulAlgalBloom". Default is "OceanDrift".
    export_variables : list
        List of variables to export, by default None. See PTM docs for options.
    radius : int
        Radius around each lon-lat pair, within which particles will be randomly seeded. This is used by function `seed_elements`.
    radius_type : str
        If 'gaussian' (default), the radius is the standard deviation in x-y-directions. If 'uniform', elements are spread evenly and always inside a circle with the given radius. This is used by function `seed_elements`.
    current_uncertainty : float
        Add gaussian perturbation with this standard deviation to current components at each time step.
    wind_uncertainty : float
        Add gaussian perturbation with this standard deviation to wind components at each time step.
    use_auto_landmask : bool
        Set as True to use general landmask instead of that from ocean_model.
        Use for testing primarily. Default is False.
    diffusivitymodel : str
        Algorithm/source used for profile of vertical diffusivity. Environment means that diffusivity is acquired from readers or environment constants/fallback. Turned on if ``vertical_mixing==True``.
    stokes_drift : bool
        Set to True to turn on Stokes drift, by default True. This enables 3 settings in OpenDrift:

        * o.set_config('drift:use_tabularised_stokes_drift', True)
        * o.set_config('drift:tabularised_stokes_drift_fetch', '25000')  # default
        * o.set_config('drift:stokes_drift_profile', 'Phillips')  # default

        The latter two configurations are not additionally set in OpenDriftModel since they are already the default once stokes_drift is True.
    mixed_layer_depth : float
        Fallback value for ocean_mixed_layer_thickness if not available from any reader. This is used in the calculation of vertical diffusivity.
    coastline_action : str, optional
        Action to perform if a drifter hits the coastline, by default "previous". Options
        are 'stranding', 'previous'.
    seafloor_action : str, optional
        Action to perform if a drifter hits the seafloor, by default "deactivate". Options
        are 'deactivate', 'previous', 'lift_to_seafloor'.
    max_speed : int
        Typical maximum speed of elements, used to estimate reader buffer size.
    wind_drift_depth : float
        The direct wind drift (windage) is linearly decreasing from the surface value (wind_drift_factor) until 0 at this depth.
    vertical_mixing_timestep : float
        Time step used for inner loop of vertical mixing.
    interpolator_filename : Optional[Union[pathlib.Path,str]], optional
        Filename to save interpolators to, by default None. The full path should be given, but no suffix.
        Use this to either read from an existing file at a non-default location or to save to a
        non-default location. If None and use_cache==True, the filename is set to a built-in name to an
        `appdirs` cache directory.
    plots : dict, optional
        Dictionary of plot names, their filetypes, and any kwargs to pass along, by default None.
        Available plot names are "spaghetti", "animation", "oil", "all".


    object_type: str = config_model["object_type"]["default"],
        Leeway object category for this simulation.

    diameter : float
        Seeding value of diameter. For LarvalFish simulation.
    neutral_buoyancy_salinity : float
        Seeding value of neutral_buoyancy_salinity. For LarvalFish simulation.
    stage_fraction : float
        Seeding value of stage_fraction. For LarvalFish simulation.
    hatched : float
        Seeding value of hatched. For LarvalFish simulation.
    length : float
        Seeding value of length. For LarvalFish simulation.
    weight : float
        Seeding value of weight. For LarvalFish simulation.

    oil_type : str
        Oil type to be used for the simulation, from the NOAA ADIOS database. For OpenOil simulation.
    m3_per_hour : float
        The amount (volume) of oil released per hour (or total amount if release is instantaneous). For OpenOil simulation.
    oil_film_thickness : float
        Seeding value of oil_film_thickness. For OpenOil simulation.
    droplet_size_distribution : str
        Droplet size distribution used for subsea release. For OpenOil simulation.
    droplet_diameter_mu : float
        The mean diameter of oil droplet for a subsea release, used in normal/lognormal distributions. For OpenOil simulation.
    droplet_diameter_sigma : float
        The standard deviation in diameter of oil droplet for a subsea release, used in normal/lognormal distributions. For OpenOil simulation.
    droplet_diameter_min_subsea : float
        The minimum diameter of oil droplet for a subsea release, used in uniform distribution. For OpenOil simulation.
    droplet_diameter_max_subsea : float
        The maximum diameter of oil droplet for a subsea release, used in uniform distribution. For OpenOil simulation.
    emulsification : bool
        Surface oil is emulsified, i.e. water droplets are mixed into oil due to wave mixing, with resulting increase of viscosity. For OpenOil simulation.
    dispersion : bool
        Oil is removed from simulation (dispersed), if entrained as very small droplets. For OpenOil simulation.
    evaporation : bool
        Surface oil is evaporated. For OpenOil simulation.
    update_oilfilm_thickness : bool
        Oil film thickness is calculated at each time step. The alternative is that oil film thickness is kept constant with value provided at seeding. For OpenOil simulation.
    biodegradation : bool
        Oil mass is biodegraded (eaten by bacteria). For OpenOil simulation.
    """

    def __init__(self, **kwargs: dict) -> None:
        """Initialize OpenDriftModel."""
        # Initialize the parent class
        # This sets up the logger and ParticleTrackingState and SetupOutputFiles.
        super().__init__(**kwargs)

        # OpenDriftConfig is a subclass of TheManagerConfig so it knows about all the
        # TheManagerConfig parameters. TheManagerConfig is run with OpenDriftConfig.
        drift_model = kwargs.get("drift_model", "OceanDrift")
        assert isinstance(drift_model, str)
        if "drift_model" in kwargs:
            del kwargs["drift_model"]
        self.config: OpenDriftConfig = open_drift_mapper[drift_model](**kwargs)

        # copy output_file from files to config. This isn't strictly necessary and it
        # seems like there is a better way to do this but do it for now to avoid
        # confusion between self.files.output_file and self.config.output_file otherwise
        # being different
        self.config.output_file = self.files.output_file

        # Note that you can see configuration possibilities for a given model with
        # o.list_configspec()
        # You can check the metadata for a given configuration with (min/max/default/type)
        # o.get_configspec('vertical_mixing:timestep')
        # You can check required variables for a model with
        # o.required_variables

        # TODO: streamline this
        self.checked_plot = False

    def _create_opendrift_model_object(self) -> None:
        """Create the OpenDrift model object."""

        # do this right away so I can query the object
        # we don't actually input output_format here because we first output to netcdf, then
        # resave as parquet after adding in extra config
        # TODO: should drift_model be instantiated in OpenDriftConfig or here?
        log_level = logger.level
        iomodule = self.config.output_format
        if iomodule == "both":
            iomodule = (
                OutputFormatEnum.netcdf
            )  # first output netcdf, then additionally output parquet at the end
        if self.config.drift_model == "Leeway":
            from opendrift.models.leeway import Leeway

            o = Leeway(loglevel=log_level, iomodule=iomodule)

        elif self.config.drift_model == "OceanDrift":
            from opendrift.models.oceandrift import OceanDrift

            o = OceanDrift(loglevel=log_level, iomodule=iomodule)

        elif self.config.drift_model == "LarvalFish":
            from opendrift.models.larvalfish import LarvalFish

            o = LarvalFish(loglevel=log_level, iomodule=iomodule)

        elif self.config.drift_model == "OpenOil":
            from opendrift.models.openoil import OpenOil

            o = OpenOil(loglevel=log_level, iomodule=iomodule, weathering_model="noaa")

        elif self.config.drift_model == "HarmfulAlgalBloom":
            from opendrift.models.harmfulalgalbloom import HarmfulAlgalBloom

            o = HarmfulAlgalBloom(loglevel=log_level, iomodule=iomodule)

        else:
            raise ValueError(
                f"Drifter model {self.config.drift_model} is not recognized."
            )

        self.o = o

    def _update_od_config_from_this_config(self) -> None:
        """Update OpenDrift's config values with OpenDriftConfig and TheManagerConfig.

        Update the default value in OpenDrift's config dict with the
        config value from OpenDriftConfig, TheManagerConfig, OceanModelConfig, and SetupOutputFiles.

        This uses the metadata key "od_mapping" to map from the PTM parameter
        name to the OpenDrift parameter name.
        """

        base_models_to_check = [self.config, self.files, self.config.ocean_model_config]
        for base_model in base_models_to_check:
            for field_name, field in base_model.model_fields.items():
                if (
                    field.json_schema_extra is None
                    or "od_mapping" not in field.json_schema_extra
                ):
                    continue

                od_key = field.json_schema_extra["od_mapping"]
                if od_key not in self.o._config:
                    continue

                # want the string representation of only this one used
                if od_key == "seed:oil_type":
                    # for oil_type, copy the oil name into the OpenDrift config
                    field_value = OIL_ID_TO_NAME[getattr(base_model, field_name)]
                # for others use value
                else:
                    field_value = getattr(base_model, field_name)
                    # if isinstance(field_value, Enum):
                    #     field_value = field_value.value

                self.o._config[od_key]["value"] = field_value

    def _modify_opendrift_model_object(self) -> None:
        """Modify the OpenDrift model object."""
        if self.config.stokes_drift:
            self.o.set_config("drift:use_tabularised_stokes_drift", True)
            # self.o.set_config('drift:tabularised_stokes_drift_fetch', '25000')  # default
            # self.o.set_config('drift:stokes_drift_profile', 'Phillips')  # default

        # If 2D surface simulation (and not Leeway since not available), truncate model output below 0.5 m
        if (
            not self.config.do3D
            and self.config.z == 0
            and self.config.drift_model != "Leeway"
        ):
            self.o.set_config("drift:truncate_ocean_model_below_m", 0.5)
            logger.debug("Truncating model output below 0.5 m.")

        # If 2D simulation (and not Leeway since not available), turn off vertical advection
        if not self.config.do3D and self.config.drift_model != "Leeway":
            self.o.set_config("drift:vertical_advection", False)
            logger.debug("Disabling vertical advection.")

        # If 3D simulation, turn on vertical advection
        if self.config.do3D:
            self.o.set_config("drift:vertical_advection", True)
            logger.debug("do3D is True so turning on vertical advection.")

        # # Assign oil_type for OpenOil simulation by id to be unique since oil names are not unique
        # # WARHING! o.set_oiltype_by_id doesn't appear to actually set the oil type (only the name, not the config)
        # if self.config.drift_model == "OpenOil":
        #     oil_type_id = OilTypeEnum(self.config.oil_type).name
        #     self.o.set_oiltype_by_id(oil_type_id)
        #     logger.debug(f"Setting oil type {self.config.oil_type} by id {oil_type_id}.")

        # TODO: what if I change the order of when config is updated with this method, then
        # does it impact that i sed oil type by id?

    def _setup_for_simulation(self) -> None:
        """Set up the simulation.
        This is run before the simulation starts.
        """

        self.logger_config.merge_with_opendrift_log()
        self._create_opendrift_model_object()
        self._update_od_config_from_this_config()
        self._modify_opendrift_model_object()

    def _add_reader(self, **kwargs: Any) -> None:
        """Add a reader to the OpenDrift model.

        Parameters
        ----------
        ds : xr.Dataset, optional
            Previously-opened Dataset containing ocean model output, if user wants to input
            unknown reader information.
        name : str, optional
            If ds is input, user can also input name of ocean model, otherwise will be called "user_input".
        """
        # Extract specific parameters from kwargs
        ds = kwargs.get("ds", None)
        name = kwargs.get("name", None)
        oceanmodel_lon0_360 = kwargs.get("oceanmodel_lon0_360", False)
        standard_name_mapping = kwargs.get("standard_name_mapping", None)

        if not self.state.has_run_setup:
            self._setup_for_simulation()

        if ds is None:
            ds = self.config.ocean_model_simulation.open_dataset(
                drop_vars=self.config.drop_vars
            )

        # don't need the following currently if using ocean_model_local since the kerchunk file is already
        # narrowed to the simulation size
        if not self.config.ocean_model_local:
            assert self.config.start_time is not None
            assert self.config.end_time is not None
            ds = narrow_dataset_to_simulation_time(
                ds, self.config.start_time, self.config.end_time
            )
            logger.debug("Narrowed model output to simulation time")

        ds = apply_known_ocean_model_specific_changes(
            ds, self.config.ocean_model_config.name, self.config.use_static_masks
        )

        if (
            self.config.ocean_model_config.name not in ocean_model_registry.all()
            and self.config.ocean_model_config.name != "test"
        ):
            ds = apply_user_input_ocean_model_specific_changes(
                ds, self.config.use_static_masks
            )

        self.ds = ds

        reader = reader_ROMS_native.Reader(
            filename=ds,
            name=self.config.ocean_model_config.name,
            standard_name_mapping=self.config.ocean_model_config.standard_name_mapping,
            save_interpolator=self.config.save_interpolator,
            interpolator_filename=self.config.interpolator_filename,
        )

        self.o.add_reader([reader])
        self.reader = reader
        # can find reader at manager.o.env.readers[self.ocean_model.name]

    @property
    def seed_kws(self) -> dict:
        """Gather seed input kwargs.

        This could be run more than once.
        """

        already_there = [
            "seed:number",
            "seed:z",
            "seed:seafloor",
            "seed:droplet_diameter_mu",
            "seed:droplet_diameter_min_subsea",
            "seed:droplet_size_distribution",
            "seed:droplet_diameter_sigma",
            "seed:droplet_diameter_max_subsea",
            "seed:object_type",
            "seed:ocean_only",
            "seed_flag",
            "drift:use_tabularised_stokes_drift",
            "drift:vertical_advection",
            "drift:truncate_ocean_model_below_m",
        ]

        time: float | datetime | list[float] | str | None
        if self.config.start_time_end is not None:
            # time can be a list to start drifters linearly in time
            time = [
                pd.Timestamp(self.config.start_time).to_pydatetime(),
                pd.Timestamp(self.config.start_time_end).to_pydatetime(),
            ]
        elif self.config.start_time is not None:
            time = self.config.start_time
            # time = self.config.start_time.to_pydatetime()
        else:
            time = None

        if self.config.seed_flag == "geojson":
            # geojson needs string representation of time
            time = (
                self.config.start_time.isoformat()
                if self.config.start_time is not None
                else None
            )

        _seed_kws = {
            "time": time,
            "z": self.config.z,
        }

        # update seed_kws with drift_model-specific seed parameters
        seedlist = {
            k: v["value"] for k, v in self.o.get_configspec(prefix="seed").items()
        }
        seedlist = {
            one.replace("seed:", ""): two
            for one, two in seedlist.items()
            if one not in already_there
        }
        _seed_kws.update(seedlist)

        if self.config.seed_flag == "elements":
            # add additional seed parameters
            _seed_kws.update(
                {
                    "lon": self.config.lon,
                    "lat": self.config.lat,
                    "radius": self.config.radius,
                    "radius_type": self.config.radius_type,
                }
            )
        elif self.config.seed_flag == "geojson":
            # add additional seed parameters
            _seed_kws.update(
                {
                    "radius": self.config.radius,
                    "radius_type": self.config.radius_type,
                }
            )

        self._seed_kws = _seed_kws
        return self._seed_kws

    def _seed(self) -> None:
        """Actually seed drifters for model."""

        if self.config.seed_flag == "elements":
            self.o.seed_elements(**self.seed_kws)

        elif self.config.seed_flag == "geojson":

            # # geojson needs string representation of time
            # self.seed_kws["time"] = self.config.start_time.isoformat()
            self.config.geojson["properties"] = self.seed_kws  # type: ignore
            json_string_dumps = json.dumps(self.config.geojson)
            self.o.seed_from_geojson(json_string_dumps)

        else:
            raise ValueError(f"seed_flag {self.config.seed_flag} not recognized.")

        self.initial_drifters = self.o.elements_scheduled

    def _run(self) -> None:
        """Run the drifters!"""

        # add input config to model config
        self.o.metadata_dict.update(self.config.model_dump())
        self.o.metadata_dict.update(self.files.model_dump())

        # actually run
        self.o.run(
            time_step=self.config.time_step,
            time_step_output=self.config.time_step_output,
            steps=self.config.steps,
            export_variables=self.config.export_variables,
            outfile=str(self.files.output_file),
        )

        # if output format is both, also save as parquet
        if self.config.output_format == "both":
            assert self.files.output_file is not None
            parquet_file = Path(self.files.output_file).with_suffix(".parquet")

            # Open the netCDF as an xarray.Dataset
            ds = xr.open_dataset(self.files.output_file)

            # Expect dims (trajectory, time)
            # Convert to a tidy DataFrame: columns = trajectory, time, variables
            df = ds.to_dataframe().reset_index()

            # # Optional: drop rows with NaN lon as “unseeded”/invalid
            # if "lon" in df.columns:
            #     df = df[~df["lon"].isna()]

            # Ensure time is datetime64[ns]
            df["time"] = pd.to_datetime(df["time"])

            # Save to parquet
            df.to_parquet(parquet_file, engine="fastparquet")
            logger.info(f"Also saved output to parquet file {parquet_file}.")

        # plot if requested
        if self.config.plots:
            assert isinstance(self.files.output_file, Path)
            # return plots because now contains the filenames for each plot
            self.config.plots = make_plots(
                self.config.plots,
                self.o,
                str(self.files.output_file).split(".")[0],
                self.config.drift_model,
            )

            # convert plots dict into string representation to save in output file attributes
            # https://github.com/pydata/xarray/issues/1307
            self.config.plots = repr(self.config.plots)

    def all_export_variables(self) -> list:
        """Output list of all possible export variables."""

        if not self.state.has_run_setup:
            self.setup_for_simulation()
            logger.debug(
                "Running setup for simulation so that full export variable list is available."
            )

        vars = (
            list(self.o.elements.variables.keys())
            + ["trajectory", "time"]
            + list(self.o.required_variables.keys())
        )

        return vars

    def export_variables(self) -> list:
        """Output list of all actual export variables."""

        return self.o.export_variables

    def reader_metadata(self, key: str) -> dict:
        """allow manager to query reader metadata."""

        if not self.state.has_added_reader:
            raise ValueError("reader has not been added yet.")
        return self.o.env.readers[self.config.ocean_model_config.name].__dict__[key]



from particle_tracking_manager.config_logging import LoggerConfig


def test_log_level():
    """Test values of parameters being input."""

    m = LoggerConfig(
        log_level="DEBUG",
    )
    assert m.log_level == "DEBUG"



"""From Copilot"""

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pydantic import ValidationError

import particle_tracking_manager as ptm

from particle_tracking_manager.models.opendrift.enums.oil_types import OIL_ID_TO_NAME
from particle_tracking_manager.models.opendrift.opendrift import OpenDriftModel


ds = xr.Dataset(
    data_vars={
        "u": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "v": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "w": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "salt": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "temp": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "wetdry_mask_rho": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "mask_rho": (("Y", "X"), np.zeros((2, 3))),
        "Uwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Vwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Cs_r": (("Z"), np.linspace(-1, 0, 3)),
        "hc": 16,
    },
    coords={
        "ocean_time": ("ocean_time", [0, 1], {"units": "seconds since 1970-01-01"}),
        "s_rho": (("Z"), np.linspace(-1, 0, 3)),
        "lon_rho": (("Y", "X"), np.array([[1, 2, 3], [1, 2, 3]])),
        "lat_rho": (("Y", "X"), np.array([[1, 1, 1], [2, 2, 2]])),
    },
)
ds_info = dict(
    lon_min=1, lon_max=3, lat_min=1, lat_max=2, start_time_model=0, end_time_fixed=1
)

ptm.config_ocean_model.register_on_the_fly(ds_info)

# ocean_model_local is False because otherwise it requires a `kerchunk_func_str`
seed_kws = dict(lon=2, lat=1.5, start_time=0, time_step=0.01, ocean_model_local=False)


def test_start_time_tz():
    """Check start time timezone is removed."""

    m = OpenDriftModel(
        duration="1s",
        start_time="1970-01-01T00:00Z",
        lon=2,
        lat=1.5,
        time_step=0.01,
        ocean_model="ONTHEFLY",
        ocean_model_local=False,
    )
    m.add_reader(ds=ds)
    assert m.config.start_time == pd.Timestamp("1970-01-01 00:00:00")
    assert m.config.end_time == pd.Timestamp("1970-01-01 00:00:01")

    m = OpenDriftModel(
        duration="1s",
        start_time=(pd.Timestamp("1970-01-01T00:00") - pd.Timedelta("5h")).tz_localize(
            "US/Eastern"
        ),
        lon=2,
        lat=1.5,
        time_step=0.01,
        ocean_model="ONTHEFLY",
        ocean_model_local=False,
    )
    m.add_reader(ds=ds)
    assert m.config.start_time == pd.Timestamp("1970-01-01 00:00:00")
    assert m.config.end_time == pd.Timestamp("1970-01-01 00:00:01")


def test_drop_vars_do3D_true():
    m = OpenDriftModel(
        drift_model="OceanDrift",
        do3D=True,
        duration="1s",
        ocean_model="ONTHEFLY",
        **seed_kws
    )
    # drop variables manually using drop_vars from config to check behavior
    m.add_reader(ds=ds.drop_vars(m.config.drop_vars, errors="ignore"))
    assert m.reader.variables == [
        "x_sea_water_velocity",
        "y_sea_water_velocity",
        "upward_sea_water_velocity",
        "land_binary_mask",
        "x_wind",
        "y_wind",
        "wind_speed",
        "sea_water_speed",
    ]


def test_drop_vars_do3D_false_use_static_masks():
    m = OpenDriftModel(
        drift_model="OceanDrift",
        use_static_masks=True,
        duration="1s",
        ocean_model="ONTHEFLY",
        **seed_kws
    )
    # drop variables manually using drop_vars from config to check behavior
    m.add_reader(ds=ds.drop_vars(m.config.drop_vars, errors="ignore"))
    assert m.reader.variables == [
        "x_sea_water_velocity",
        "y_sea_water_velocity",
        "land_binary_mask",
        "x_wind",
        "y_wind",
        "wind_speed",
        "sea_water_speed",
    ]
    assert "mask_rho" in m.reader.Dataset.data_vars
    assert "wetdry_mask_rho" not in m.reader.Dataset.data_vars


def test_drop_vars_no_wind():
    m = OpenDriftModel(
        drift_model="OceanDrift",
        duration="1s",
        stokes_drift=False,
        wind_drift_factor=0,
        wind_uncertainty=0,
        vertical_mixing=False,
        ocean_model="ONTHEFLY",
        **seed_kws
    )
    # drop variables manually using drop_vars from config to check behavior
    m.add_reader(ds=ds.drop_vars(m.config.drop_vars, errors="ignore"))
    assert m.reader.variables == [
        "x_sea_water_velocity",
        "y_sea_water_velocity",
        "land_binary_mask",
        "sea_water_speed",
    ]


def test_Leeway():

    m = OpenDriftModel(
        drift_model="Leeway",
        object_type=">PIW, scuba suit (face up)",
        stokes_drift=False,
        steps=1,
        ocean_model="ONTHEFLY",
        **seed_kws
    )
    m.setup_for_simulation()  # creates m.o
    # assert "wind_drift_factor" not in m.config
    assert not hasattr(m.config, "wind_drift_factor")
    assert "seed:wind_drift_factor" not in m.o.get_configspec().keys()


def test_LarvalFish_add_reader():
    m = OpenDriftModel(
        drift_model="LarvalFish",
        do3D=True,
        duration="1s",
        ocean_model="ONTHEFLY",
        **seed_kws
    )
    m.add_reader(ds=ds.drop_vars(m.config.drop_vars, errors="ignore"))
    assert m.reader.variables == [
        "x_sea_water_velocity",
        "y_sea_water_velocity",
        "upward_sea_water_velocity",
        "sea_water_salinity",
        "sea_water_temperature",
        "land_binary_mask",
        "x_wind",
        "y_wind",
        "wind_speed",
        "sea_water_speed",
    ]


def test_LarvalFish_seeding():
    """Make sure special seed parameter comes through"""

    m = OpenDriftModel(
        drift_model="LarvalFish",
        lon=-151,
        lat=60,
        do3D=True,
        hatched=1,
        stage_fraction=1,  # value has to be a number but won't be used since hatched==1
        start_time="2022-01-01T00:00:00",
        use_auto_landmask=True,
        steps=1,
        vertical_mixing=True,
        wind_drift_factor=0,
        wind_drift_depth=0,
    )
    m.setup_for_simulation()  # creates m.o
    assert m.o._config["seed:hatched"]["value"] == 1


def test_OpenOil_seeding():
    """Make sure special seed parameters comes through"""
    oil_type = "AD00010"
    m = OpenDriftModel(
        drift_model="OpenOil",
        lon=-151,
        lat=60,
        do3D=True,
        start_time="2023-01-01T00:00:00",
        use_auto_landmask=True,
        m3_per_hour=5,
        droplet_diameter_max_subsea=0.1,
        droplet_diameter_min_subsea=0.01,
        droplet_diameter_mu=0.01,
        droplet_size_distribution="normal",
        droplet_diameter_sigma=0.9,
        oil_film_thickness=5,
        oil_type=oil_type,
        # oil_type="GN00002",
        steps=1,
    )
    m.setup_for_simulation()  # creates m.o

    # m.o.set_config("environment:constant:x_wind", -1)
    # m.o.set_config("environment:constant:y_wind", -1)
    # m.o.set_config("environment:constant:x_sea_water_velocity", -1)
    # m.o.set_config("environment:constant:y_sea_water_velocity", -1)
    # m.o.set_config("environment:constant:sea_water_temperature", 15)
    # m.seed()

    # to check impact of m3_per_hour: mass_oil for m3_per_hour of 1 * 5
    # assert np.allclose(m.o.elements_scheduled.mass_oil, 0.855 * 5)  # i'm getting different answers local vs github actiosn
    assert m.o._config["seed:m3_per_hour"]["value"] == 5
    assert m.o._config["seed:droplet_diameter_max_subsea"]["value"] == 0.1
    assert m.o._config["seed:droplet_diameter_min_subsea"]["value"] == 0.01
    assert m.o._config["seed:droplet_diameter_mu"]["value"] == 0.01
    assert m.o._config["seed:droplet_size_distribution"]["value"] == "normal"
    assert m.o._config["seed:droplet_diameter_sigma"]["value"] == 0.9
    # assert m.o.elements_scheduled.oil_film_thickness == 5
    assert (
        m.o._config["seed:oil_type"]["value"] == OIL_ID_TO_NAME[oil_type]
    )  # don't use ID because it is stripped off


def test_OpenOil_oil_type_id():
    """Make sure oil type entered as id works"""
    oil_type = "AD00010"
    m = OpenDriftModel(drift_model="OpenOil", oil_type=oil_type, steps=1)


@pytest.mark.slow
def test_OpenOil_vertical_mixing():
    m = ptm.OpenDriftModel(
        drift_model="OpenOil",
        steps=1,
        oil_type="AD00010",
        do3D=True,
        vertical_mixing=False,
    )
    m.setup_for_simulation()
    assert not m.o.get_config("drift:vertical_mixing")


@pytest.mark.slow
def test_OpenOil_all_oils_exact_match():
    """Make sure that oils in PTM exactly match those in OpenDrift."""
    import opendrift.models.openoil.adios.dirjs as dirjs

    schema = ptm.OpenOilModelConfig.model_json_schema()
    ptm_oils = {
        oil["const"]: oil["title"] for oil in schema["properties"]["oil_type"]["oneOf"]
    }
    od_oils = {oil.id: oil.name for oil in dirjs.oils(limit=1300)}

    # OpenDrift has 5 extra Norwegian oils that are not in PTM so we remove them for this test
    extra_od_oils = ["NO00167", "NO00168", "NO00169", "NO00170", "NO00171"]
    for extra_oil in extra_od_oils:
        od_oils.pop(extra_oil, None)

    assert ptm_oils == od_oils


def test_HarmfulAlgalBloom_seeding():
    """Make sure special parameter comes through"""

    m = OpenDriftModel(
        drift_model="HarmfulAlgalBloom",
        species_type="PN",
        temperature_death_min=1,
        lon=-151,
        lat=60,
        start_time="2022-01-01T00:00:00",
        use_auto_landmask=True,
        steps=1,
    )
    m.setup_for_simulation()  # creates m.o
    assert m.o._config["hab:temperature_death_min"]["value"] == 1


def test_wind_drift():
    """Make sure changed wind drift numbers comes through"""

    m = OpenDriftModel(
        drift_model="OceanDrift",
        lon=-151,
        lat=60,
        do3D=True,
        wind_drift_factor=0.05,
        wind_drift_depth=0.5,
        start_time="2023-01-01T00:00:00",
        use_auto_landmask=True,
        steps=1,
    )
    m.setup_for_simulation()  # creates m.o
    assert m.o._config["seed:wind_drift_factor"]["value"] == 0.05
    assert m.o._config["drift:wind_drift_depth"]["value"] == 0.5


def test_plots_linecolor():
    # since export_variables are all included by default now, I am not testing this
    # this should error if user inputs some export_variables, which
    # changes the default from returning all variables to just those
    # selected plus a short list of required variables
    # with pytest.raises(ValueError):
    #     m = OpenDriftModel(
    #         drift_model="OceanDrift",
    #         plots={"spaghetti": {"linecolor": "x_wind"}},
    #         export_variables=[],
    #         steps=1
    #     )

    m = OpenDriftModel(
        drift_model="OceanDrift", plots={"spaghetti": {"linecolor": "x_wind"}}, steps=1
    )

    m = OpenDriftModel(
        drift_model="OceanDrift",
        plots={"spaghetti": {"linecolor": "x_wind"}},
        export_variables=None,
        steps=1,
    )

    # this should work bc "z" should already be included
    m = OpenDriftModel(
        drift_model="OceanDrift", plots={"spaghetti": {"linecolor": "z"}}, steps=1
    )


def test_plots_background():
    # # this should error if user inputs some export_variables, which
    # # changes the default from returning all variables to just those
    # # selected plus a short list of required variables
    # with pytest.raises(ValueError):
    #     m = OpenDriftModel(
    #         drift_model="OceanDrift",
    #         plots={"animation": {"background": "sea_surface_height"}},
    #         export_variables=[],
    #         steps=1
    #     )

    m = OpenDriftModel(
        drift_model="OceanDrift",
        plots={"animation": {"background": "sea_surface_height"}},
        steps=1,
    )


def test_plots_oil():
    # # this should error if user inputs some export_variables, which
    # # changes the default from returning all variables to just those
    # # selected plus a short list of required variables
    # with pytest.raises(ValueError):
    #     m = OpenDriftModel(
    #         drift_model="OpenOil",
    #         plots={"oil": {"show_wind_and_current": True}},
    #         export_variables=[],
    #     )

    m = OpenDriftModel(
        drift_model="OpenOil", plots={"oil": {"show_wind_and_current": True}}, steps=1
    )

    with pytest.raises(ValidationError):
        m = OpenDriftModel(drift_model="OceanDrift", plots={"oil": {}}, steps=1)


def test_plots_property():
    # # this should error if user inputs some export_variables, which
    # # changes the default from returning all variables to just those
    # # selected plus a short list of required variables
    # with pytest.raises(ValueError):
    #     m = OpenDriftModel(
    #         drift_model="LarvalFish",
    #         do3D=True,
    #         plots={"property": {"prop": "survival"}},
    #         export_variables=["x_wind"],
    #     )

    m = OpenDriftModel(
        drift_model="LarvalFish",
        do3D=True,
        plots={"property": {"prop": "survival"}},
        steps=1,
    )


def test_plots_all():

    with pytest.raises(ValueError):
        m = OpenDriftModel(
            drift_model="OceanDrift",
            plots={
                "all": {},
                "spaghetti": {"line_color": "x_wind"},
                "animation": {"background": "sea_surface_height"},
            },
            steps=1,
        )


def test_plots_names():

    with pytest.raises(ValidationError):
        m = OpenDriftModel(
            plots={
                "random_plot_name": {},
            },
            steps=1,
        )

    with pytest.raises(ValidationError):
        m = OpenDriftModel(
            plots={
                "something_spaghetti": {},
            },
            steps=1,
        )


# @pytest.mark.slow
def test_parameter_passing():
    """make sure parameters passed into package make it to simulation runtime."""

    ts = 7 * 60  # seconds
    diffmodel = "windspeed_Sundby1983"
    use_auto_landmask = True
    vertical_mixing = True
    do3D = True

    seed_kws = dict(
        lon=-151,
        lat=60.0,
        radius=5000,
        number=100,
        start_time=datetime(2022, 9, 22, 6, 0, 0),
    )
    m = OpenDriftModel(
        use_auto_landmask=use_auto_landmask,
        time_step=ts,
        duration="P0DT0H5M0S",
        steps=None,
        diffusivitymodel=diffmodel,
        vertical_mixing=vertical_mixing,
        do3D=do3D,
        **seed_kws
    )

    m.setup_for_simulation()  # creates m.o

    # idealized simulation, provide a fake current
    m.o.set_config("environment:fallback:y_sea_water_velocity", 1)

    # check time_step across access points
    assert (
        # m.o._config["general:time_step_minutes"]["value"]  # this is not correct, don't know why
        # m.o.time_step.total_seconds()  # this is only created once run
        ts
        == m.config.time_step
        # == m.o.get_configspec()["general:time_step_minutes"]["value"]  # this is not correct, don't know why
    )

    # check diff model
    assert (
        m.o.get_configspec()["vertical_mixing:diffusivitymodel"]["value"] == diffmodel
    )

    # check use_auto_landmask coming through
    assert (
        m.o.get_configspec()["general:use_auto_landmask"]["value"] == use_auto_landmask
    )



from pathlib import Path

import pytest

from particle_tracking_manager.config_misc import SetupOutputFiles


@pytest.mark.parametrize("suffix", ["", ".nc", ".parq", ".parquet", ".txt"])
def test_log_name(suffix):
    m = SetupOutputFiles(output_file=f"outputs/newtest{suffix}")
    assert m.logfile_name == str(Path("outputs/newtest.log"))


@pytest.mark.parametrize("format, suffix", [("netcdf", ".nc"), ("parquet", ".parquet")])
def test_output_file(format, suffix):
    """Make sure output file suffix matches the chosen format"""

    m = SetupOutputFiles(output_format=format)
    assert m.output_file.suffix == suffix



"""OpenDrift HAB species types enum definition."""

from enum import Enum

from pydantic import BaseModel, ConfigDict, Field


class HABSpeciesTypeEnum(str, Enum):
    """Harmful Algal Bloom species types supported by OpenDrift."""

    PN = "PN"
    AX = "AX"
    DP = "DP"
    custom = "custom"


HAB_SPECIES_LABELS = {
    HABSpeciesTypeEnum.PN: "Pseudo nitzschia",
    HABSpeciesTypeEnum.AX: "Alexandrium",
    HABSpeciesTypeEnum.DP: "Dinophysis",
    HABSpeciesTypeEnum.custom: "Custom species (manual parameters)",
}


class VerticalBehaviorEnum(str, Enum):
    """Enum for vertical behavior types."""

    none = "none"
    band = "band"
    diel_band = "diel_band"


class HABParameters(BaseModel):
    """Harmful Algal Bloom species parameters for OpenDrift."""

    model_config = ConfigDict(extra="forbid")

    temperature_death_min: float = Field(
        description="Minimum temperature for living. Below this temperature, cells have high mortality rate. Between this and temperature_viable_min, cells have no growth.",
        title="Cell death below this temperature",
        ge=-5.0,
        le=40.0,
        default=3.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_death_min",
            "ptm_level": 2,
        },
    )

    temperature_death_max: float = Field(
        description="Maximum temperature for living. Above this temperature, cells have high mortality rate. Between temperature_viable_max and this parameter, cells have no growth.",
        title="Cell death above this temperature",
        ge=-5.0,
        le=40.0,
        default=22.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_death_max",
            "ptm_level": 2,
        },
    )

    temperature_pref_min: float = Field(
        description="Minimum temperature for preferred temperature range; cells have regular growth.",
        title="Minimum preferred temperature",
        ge=-5.0,
        le=40.0,
        default=10.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_pref_min",
            "ptm_level": 2,
        },
    )

    temperature_pref_max: float = Field(
        description="Maximum temperature for preferred temperature range; cells have regular growth.",
        title="Maximum preferred temperature",
        ge=-5.0,
        le=40.0,
        default=20.0,
        json_schema_extra={
            "units": "degrees",
            "od_mapping": "hab:temperature_pref_max",
            "ptm_level": 2,
        },
    )

    salinity_death_min: float = Field(
        description="Minimum salinity for living. Below this salinity, cells have high mortality rate. Between this and salinity_viable_min, cells have no growth.",
        title="Cell death below this salinity",
        ge=0.0,
        le=50.0,
        default=25.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_death_min",
            "ptm_level": 2,
        },
    )

    salinity_death_max: float = Field(
        description="Maximum salinity for living. Above this salinity, cells have high mortality rate. Between salinity_viable_max and this parameter, cells have no growth.",
        title="Cell death above this salinity",
        ge=0.0,
        le=50.0,
        default=36.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_death_max",
            "ptm_level": 2,
        },
    )

    salinity_pref_min: float = Field(
        description="Minimum salinity for preferred salinity range; cells have regular growth.",
        title="Minimum preferred salinity",
        ge=0.0,
        le=50.0,
        default=30.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_pref_min",
            "ptm_level": 2,
        },
    )

    salinity_pref_max: float = Field(
        description="Maximum salinity for preferred salinity range; cells have regular growth.",
        title="Maximum preferred salinity",
        ge=0.0,
        le=50.0,
        default=40.0,
        json_schema_extra={
            "units": "psu",
            "od_mapping": "hab:salinity_pref_max",
            "ptm_level": 2,
        },
    )

    mortality_rate_high: float = Field(
        description="Rate of mortality applied below temperature_death_min and above temperature_death_max.",
        title="High mortality rate",
        ge=0.0,
        le=10.0,
        default=1.0,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:mortality_rate_high",
            "ptm_level": 2,
        },
    )

    mortality_rate_medium: float = Field(
        description="Rate of mortality applied in edge conditions defined by temperature_death_min, temperature_death_max, salinity_death_min, salinity_death_max, temperature_pref_min, temperature_pref_max, salinity_pref_min, and salinity_pref_max.",
        title="Medium mortality rate",
        ge=0.0,
        le=10.0,
        default=0.5,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:mortality_rate_medium",
            "ptm_level": 2,
        },
    )

    mortality_rate_low: float = Field(
        description="Rate of mortality applied in preferred conditions defined by temperature_pref_min, temperature_pref_max, salinity_pref_min, and salinity_pref_max.",
        title="Low mortality rate",
        ge=0.0,
        le=10.0,
        default=0.1,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:mortality_rate_low",
            "ptm_level": 2,
        },
    )

    growth_rate_high: float = Field(
        description="Rate of growth applied in preferred conditions defined by temperature_pref_min, temperature_pref_max, salinity_pref_min, and salinity_pref_max.",
        title="High growth rate",
        ge=0.0,
        le=10.0,
        default=0.7,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:growth_rate_high",
            "ptm_level": 2,
        },
    )

    growth_rate_medium: float = Field(
        description="Rate of growth applied in edge conditions defined by temperature_death_min, temperature_death_max, salinity_death_min, salinity_death_max, temperature_pref_min, temperature_pref_max, salinity_pref_min, and salinity_pref_max.",
        title="Medium growth rate",
        ge=0.0,
        le=10.0,
        default=0.2,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:growth_rate_medium",
            "ptm_level": 2,
        },
    )

    growth_rate_low: float = Field(
        description="Rate of growth applied below viable conditions defined by temperature_death_min, temperature_death_max, salinity_death_min, and salinity_death_max.",
        title="Low growth rate",
        ge=0.0,
        le=10.0,
        default=0.0,
        json_schema_extra={
            "units": "days^-1",
            "od_mapping": "hab:growth_rate_low",
            "ptm_level": 2,
        },
    )

    vertical_behavior: VerticalBehaviorEnum = Field(
        description="Vertical behavior: no active movement, fixed band, or diel band migration.",
        title="Vertical behavior",
        default="none",
        json_schema_extra={
            "od_mapping": "hab:vertical_behavior",
            "ptm_level": 2,
        },
    )

    swim_speed: float = Field(
        description="Maximum active vertical swimming speed (m/s).",
        title="Swim speed",
        ge=0.0,
        le=100.0,
        default=0.001,
        json_schema_extra={
            "units": "m/s",
            "od_mapping": "hab:swim_speed",
            "ptm_level": 2,
        },
    )

    band_center_depth: float = Field(
        description="Target center of preferred depth band (m, negative down).",
        title="Band center depth",
        ge=-10000.0,
        le=0.0,
        default=-10.0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "hab:band_center_depth",
            "ptm_level": 2,
        },
    )

    band_half_width: float = Field(
        description="Half-width of preferred depth band (m).",
        title="Band half-width",
        ge=0.0,
        le=5000.0,
        default=5.0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "hab:band_half_width",
            "ptm_level": 2,
        },
    )

    diel_day_depth: float = Field(
        description="Target depth during daytime (m, negative).",
        title="Diel day depth",
        ge=-10000.0,
        le=0.0,
        default=-20.0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "hab:diel_day_depth",
            "ptm_level": 2,
        },
    )

    diel_night_depth: float = Field(
        description="Target depth during nighttime (m, negative).",
        title="Diel night depth",
        ge=-10000.0,
        le=0.0,
        default=-5.0,
        json_schema_extra={
            "units": "m",
            "od_mapping": "hab:diel_night_depth",
            "ptm_level": 2,
        },
    )


# Species default for HarmfulAlgalBloom model
SPECIES_HAB_DEFAULTS: dict[HABSpeciesTypeEnum, HABParameters] = {
    HABSpeciesTypeEnum.PN: HABParameters(
        temperature_death_min=2.0,
        temperature_death_max=22.0,
        temperature_pref_min=10.0,
        temperature_pref_max=18.0,
        salinity_death_min=22.0,
        salinity_death_max=36.0,
        salinity_pref_min=29.0,
        salinity_pref_max=34.0,
        mortality_rate_high=1.0,
        mortality_rate_medium=0.5,
        mortality_rate_low=0.1,
        growth_rate_high=0.8,  # day⁻¹  (within 0.4–1.04 observed)
        growth_rate_medium=0.3,  # day⁻¹
        growth_rate_low=0.0,
        vertical_behavior="none",
        # these are now calculated using mixed layer depth
        # band_center_depth=-5.0,  # m
        # band_half_width=10.0,  # m   (very broad, not tight band)
        swim_speed=0.0,
    ),
    HABSpeciesTypeEnum.AX: HABParameters(
        temperature_death_min=5.0,
        temperature_death_max=22.0,
        temperature_pref_min=8.0,
        temperature_pref_max=14.0,
        salinity_death_min=15.0,
        salinity_death_max=35.0,
        salinity_pref_min=25.0,
        salinity_pref_max=33.0,
        mortality_rate_high=1.0,
        mortality_rate_medium=0.5,
        mortality_rate_low=0.1,
        growth_rate_high=0.65,  # day⁻¹  (near 0.63 observed)
        growth_rate_medium=0.25,  # day⁻¹
        growth_rate_low=0.0,
        # Alexandrium defaults
        vertical_behavior="diel_band",
        swim_speed=0.0005,  # m/s
        # # # MLD 20 like in Lower Cook Inlet
        # # diel_day_depth=-6.0,  # m
        # # diel_night_depth=-26.0,  # m, this should be below the MLD
        # # # # MLD 12 like in KB
        # # # diel_day_depth=-4.0,       # m
        # # # diel_night_depth=-18.0,      # m
        # # # # If you ever use pure band: (not used for AX)
        # # # band_center_depth=-10.0,
        # band_half_width=5.0,
    ),
    HABSpeciesTypeEnum.DP: HABParameters(
        temperature_death_min=6.0,
        temperature_death_max=22.0,
        temperature_pref_min=10.0,
        temperature_pref_max=16.0,
        salinity_death_min=20.0,
        salinity_death_max=36.0,
        salinity_pref_min=28.0,
        salinity_pref_max=34.0,
        mortality_rate_high=1.0,
        mortality_rate_medium=0.5,
        mortality_rate_low=0.1,
        growth_rate_high=0.4,  # day⁻¹  (near 0.36–0.39 observed)
        growth_rate_medium=0.15,  # day⁻¹
        growth_rate_low=0.0,
        vertical_behavior="band",
        swim_speed=0.0003,  # m/s (≈ 26 m/day max vertical travel)
        # # Day: somewhat deeper, but not as deep as Alexandrium
        # diel_day_depth=-15.0,
        # # Night: a bit shallower, but not right at the surface
        # diel_night_depth=-5.0,
        # # Band settings:
        # band_center_depth=-12.5,
        # band_half_width=7.5,
    ),
    # HABSpeciesTypeEnum.custom intentionally has no entry
}


# # Other config defaults per species (z, do3D, etc.)
# SPECIES_HAB_MANAGER_DEFAULTS: dict[HABSpeciesTypeEnum, dict[str, object]] = {
#     HABSpeciesTypeEnum.PN: {"z": -1.0, "do3D": True, "vertical_mixing": True},
#     HABSpeciesTypeEnum.AX: {"z": -1.0, "do3D": True, "vertical_mixing": True},
#     HABSpeciesTypeEnum.DP: {"do3D": True, "vertical_mixing": True},
# }

# this is for the schema
_species_descriptions = {
    species.value: "Defaults: "
    + ", ".join(
        [
            f"{key}={value}"
            for key, value in SPECIES_HAB_DEFAULTS[species].model_dump().items()
        ]
    )
    for species in HABSpeciesTypeEnum
    if species != "custom"
}
_species_descriptions["custom"] = "Custom species with user-defined parameters."



"""Defines the ParticleTrackingManager class which is the base class for any Lagrangian model to inherit from."""

# Standard library imports
import logging

from abc import ABC, abstractmethod

# Third-party imports
from typing import Any, Self

# Local imports
from .config_logging import LoggerConfig
from .config_misc import ParticleTrackingState, SetupOutputFiles
from .config_the_manager import TheManagerConfig


logger = logging.getLogger()


class ParticleTrackingManager(ABC):
    """Manager class that controls particle tracking model.

    Parameters
    ----------
    model : str
        Name of Lagrangian model package to use for drifter tracking. Only option
        currently is "opendrift".
    lon : Optional[Union[int,float]], optional
        Central longitude for seeding drifters. If this is set, `lat` should also be set, and `geojson` should be None.
    lat : Optional[Union[int,float]], optional
        Central latitude for seeding drifters. If this is set, `lon` should also be set, and `geojson` should be None.
    geojson : Optional[dict], optional
        GeoJSON describing a polygon within which to seed drifters; must contain "geometry". If this is set, `lon` and `lat` should be None.
    start_time : Optional[str,datetime.datetime,pd.Timestamp], optional
        Start time of simulation, by default None
    start_time_end : Optional[str,datetime.datetime,pd.Timestamp], optional
        If not None, this creates a range of start times for drifters, starting with
        `start_time` and ending with `start_time_end`. Drifters will be initialized linearly
        between the two start times. Default None.
    run_forward : bool, optional
        True to run forward in time, False to run backward, by default True
    time_step : int, optional
        Time step in seconds, options >0, <86400 (1 day in seconds), by default 300.
    time_step_output : int, Timedelta, optional
        How often to output model output. Should be a multiple of time_step.
        By default 3600.
    steps : int, optional
        Number of time steps to run in simulation. Options >0.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.
    duration : Optional[datetime.timedelta], optional
        Length of simulation to run, as positive-valued timedelta object, in hours,
        such as ``timedelta(hours=48)``.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. For CLI, input duration as a pandas Timedelta
        string like "48h" for 48 hours. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.

    end_time : Optional[datetime], optional
        Datetime at which to end simulation, as positive-valued datetime object.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.

    ocean_model : Optional[str], optional
        Name of ocean model to use for driving drifter simulation, by default None.
        Use None for testing and set up. Otherwise input a string.
        Options are: "NWGOA", "CIOFS", "CIOFSOP".
        Alternatively keep as None and set up a separate reader (see example in docs).
    ocean_model_local : Optional, bool
        Set to True to use local version of known `ocean_model` instead of remote version.
    do3D : bool, optional
        Set to True to run drifters in 3D, by default False. This is overridden if
        ``surface_only==True``. If True, vertical advection and mixing are turned on with
        options for setting ``diffusivitymodel``, ``background_diffusivity``,
        ``ocean_mixed_layer_thickness``, ``vertical_mixing_timestep``. If False,
        vertical motion is disabled.
    use_static_masks : bool, optional
        Set to True to use static masks ocean_model output when ROMS wetdry masks are available, by default False.
        This is relevant for all of the available known models. If you want to use static masks
        with a user-input ocean_model, you can drop the wetdry_mask_rho etc variables from the
        dataset before inputting to PTM. Setting this to True may save computation time but
        will be less accurate, especially in the tidal flat regions of the model.
    output_file : Optional[str], optional
        Name of output file to save, by default None. If None, default is set in the model. Without any suffix.
    output_format : str, default "netcdf"
        Name of input/output module type to use for writing Lagrangian model output. Default is "netcdf".
    use_cache : bool
        Set to True to use cache for saving interpolators, by default True.
    horizontal_diffusivity : float
        Horizontal diffusivity is None by default but will be set to a grid-dependent value for known ocean_model values. This is calculated as 0.1 m/s sub-gridscale velocity that is missing from the model output and multiplied by an estimate of the horizontal grid resolution. This leads to a larger value for NWGOA which has a larger value for mean horizontal grid resolution (lower resolution). If the user inputs their own ocean_model information, they can input their own horizontal_diffusivity value. A user can use a known ocean_model and then overwrite the horizontal_diffusivity value to some value.
    log_level : str, optional
        Options are the logging input options. By default "INFO"

    """

    config: TheManagerConfig

    def __init__(self, **kwargs: dict) -> None:
        """Initialize the ParticleTrackingManager."""

        # Set up strings for the output files, which will be used in Logger setup and for all other output files.
        inputs = {
            key: kwargs[key]
            for key in ["output_file", "output_format"]
            if key in kwargs
        }
        self.files = SetupOutputFiles(**inputs)

        # Setup logging, this also contains the log_level parameter
        inputs = {key: kwargs[key] for key in ["log_level"] if key in kwargs}
        self.logger_config = LoggerConfig(**inputs)
        self.logger_config.setup_logger(logfile_name=self.files.logfile_name)
        self.state = ParticleTrackingState()

    @classmethod
    def from_config(cls, config: TheManagerConfig) -> Self:
        """Create an OpenDriftModel from a config.

        Not currently working.
        """
        return cls(**config.dict())

    def setup_for_simulation(self) -> None:
        """Set up the simulation.

        This may not be necessary to separate out but is sometimes necessary.
        """

        self._setup_for_simulation()

        # Set up state
        self.state.has_run_setup = True

    def add_reader(self, **kwargs: dict) -> None:
        """Add reader to model class."""
        self._add_reader(**kwargs)

        self.state.has_added_reader = True

    def seed(self) -> None:
        """Seed drifters."""

        if not self.state.has_added_reader:
            raise ValueError("first add reader with `manager.add_reader(**kwargs)`.")

        # run seeding function in model class
        self._seed()  # in child class

        self.state.has_run_seeding = True

    def run(self) -> None:
        """Call model run_drifters function.

        Also run some other items.
        """

        if not self.state.has_run_seeding:
            raise ValueError("first run seeding with `manager.seed()`.")

        logger.info(
            f"start_time: {self.config.start_time}, end_time: {self.config.end_time}, steps: {self.config.steps}, duration: {self.config.duration}"
        )

        self._run()  # in child class

        self.logger_config.close_loggers(logger)
        self.state.has_run = True

    def run_all(self) -> None:
        """Run all steps."""
        if not self.state.has_added_reader:
            self.add_reader()
        if not self.state.has_run_seeding:
            self.seed()
        if not self.state.has_run:
            self.run()

    @abstractmethod
    def _setup_for_simulation(self) -> None:
        """Steps to setup for specific model's simulation."""
        raise NotImplementedError("This should be implemented in the model class.")

    @abstractmethod
    def _add_reader(self, **kwargs: Any) -> None:
        """Add reader to model class."""
        raise NotImplementedError("This should be implemented in the model class.")

    @abstractmethod
    def _seed(self) -> None:
        """Seed drifters in model class."""
        raise NotImplementedError("This should be implemented in the model class.")

    @abstractmethod
    def _run(self) -> None:
        """Run drifter simulation in model class."""
        raise NotImplementedError("This should be implemented in the model class.")
